#!/usr/bin/env python3
import json
import time
import multiprocessing
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_datetime
from qualys_etl.etld_lib import etld_lib_extract_transform_load
from qualys_etl.etld_was import was_04_extract_from_qualys
import traceback


def begin_was_03_extract_controller():
    etld_lib_functions.logger.info(f"start")


def end_was_03_extract_controller():
    etld_lib_functions.logger.info(f"end")


def get_next_batch(json_file=None):
    try:
        was_batch_status = {}
        with etld_lib_config.was_open_file_compression_method(json_file, "rt", encoding='utf-8') as read_file:
            was_dict = json.load(read_file)
            was_sr_dict = {}

            if 'ServiceResponse' in was_dict.keys():
                was_sr_dict = was_dict['ServiceResponse']
            else:
                raise Exception(f"Cannot find ServiceResponse in json file: {json_file}")

            if 'SUCCESS' not in was_sr_dict['responseCode']:
                raise Exception(f"Failed Response Code in json file: {json_file}")

            if 'hasMoreRecords' not in was_sr_dict.keys():
                was_sr_dict['hasMoreRecords'] = 'false'

            if 'hasMoreRecords' in was_sr_dict.keys():
                if was_sr_dict['hasMoreRecords'] != 'true':
                    was_sr_dict['lastId'] = ""

            was_batch_status = {
                'responseCode': was_sr_dict['responseCode'],
                'count': was_sr_dict['count'],
                'hasMoreRecords': was_sr_dict['hasMoreRecords'],
                'lastId': was_sr_dict['lastId'],
            }

    except Exception as e:
        etld_lib_functions.logger.error(f"Error downloading records Exception: {e}")
        etld_lib_functions.logger.error(f"batch_status: {was_batch_status}")
        etld_lib_functions.logger.error(f"dict: {was_sr_dict}")
        etld_lib_functions.logger.error(f"file: {str(json_file)}")
        exit(1)

    return was_batch_status


def get_was_count(_: dict):
    api_call_name = _['api_call_name']
    was_extract_function_reference = _['was_extract_function_reference']
    utc_datetime = _['utc_datetime']
    cred_dict = _['cred_dict']
    dir_search_glob = _['dir_search_glob']
    # if 'last_id' in _.keys():
    #     last_id = _['last_id']
    # else:
    #     last_id = 0

    etld_lib_functions.logger.info(f"Begin {api_call_name}")
    qualys_headers_dict = {}

    etld_lib_config.remove_old_files(
        dir_path=etld_lib_config.was_extract_dir,
        dir_search_glob=dir_search_glob
    )

    file_info_dict = \
        etld_lib_config.prepare_extract_batch_file_name(
            next_batch_number_str='batch_000000',
            next_batch_date=utc_datetime,
            extract_dir=etld_lib_config.was_extract_dir,
            file_name_type=f"{api_call_name}",
            file_name_option="last_scan_date",
            file_name_option_date=etld_lib_config.was_webapp_last_scan_date,
            file_extension="json",
            compression_method=etld_lib_config.was_open_file_compression_method)

    was_extract_function_reference(
        batch_number_str=file_info_dict['next_batch_number_str'],
        qualys_headers_dict=qualys_headers_dict,
        cred_dict=cred_dict,
        file_info_dict=file_info_dict)

    etld_lib_functions.logger.info(f"End   {api_call_name}")


def get_was_data(_: dict):

    api_call_name = _['api_call_name']
    was_extract_function_reference = _['was_extract_function_reference']
    utc_datetime = _['utc_datetime']
    cred_dict = _['cred_dict']
    page_size = _['page_size']
    dir_search_glob = _['dir_search_glob']
    if 'last_id' in _.keys():
        last_id = _['last_id']
    else:
        last_id = 0

    etld_lib_config.remove_old_files(
        dir_path=etld_lib_config.was_extract_dir,
        dir_search_glob=dir_search_glob
    )

    batch_info = {'hasMoreRecords': 'true', 'lastId': last_id}
    has_more_records = 'true'
    batch_number = 0
    qualys_headers_dict = {}

    while has_more_records == 'true':
        batch_number = batch_number + 1
        batch_number_str = f'batch_{batch_number:06d}'

        # TODO currently was_webapp_last_scan_date is ignored, all apps are pulled.
        # TODO see etld_lib_config for details

        file_info_dict = \
            etld_lib_config.prepare_extract_batch_file_name(
                next_batch_number_str=batch_number_str,
                next_batch_date=utc_datetime,
                extract_dir=etld_lib_config.was_extract_dir,
                file_name_type=f"{api_call_name}",
                file_name_option="last_scan_date",
                file_name_option_date=etld_lib_config.was_webapp_last_scan_date,
                file_extension="json",
                compression_method=etld_lib_config.was_open_file_compression_method)

        was_extract_function_reference(
            last_id=str(batch_info['lastId']),
            batch_number_str=file_info_dict['next_batch_number_str'],
            qualys_headers_dict=qualys_headers_dict,
            cred_dict=cred_dict,
            file_info_dict=file_info_dict,
            page_size=page_size
        )

        batch_info = get_next_batch(json_file=file_info_dict['next_file_path'])
        has_more_records = str(batch_info['hasMoreRecords'])
        etld_lib_functions.logger.info(f"{batch_number_str} info: {batch_info}")

        has_more_records = test_system_option(
            file_path=file_info_dict['next_file_path'],
            has_more_records=has_more_records,
            number_of_files_to_extract=etld_lib_config.was_test_number_of_files_to_extract,
            test_system_flag=etld_lib_config.was_test_system_flag
        )


def test_system_option(file_path: Path, has_more_records, number_of_files_to_extract, test_system_flag):
    if test_system_flag is True:
        test_batch_number = \
            etld_lib_extract_transform_load.get_batch_number_from_filename(file_path)
        if int(test_batch_number) >= number_of_files_to_extract:
            has_more_records = '0'
    return has_more_records


def extract_was_api_data():
    begin_was_03_extract_controller()
    utc_datetime = etld_lib_datetime.get_utc_datetime_qualys_format()
    cred_dict = etld_lib_authentication_objects.qualys_authentication_obj.get_credentials_dict()
    etld_lib_config.remove_old_files(
        dir_path=etld_lib_config.was_extract_dir,
        dir_search_glob=etld_lib_config.was_extract_dir_file_search_blob,
        other_files_list=etld_lib_config.was_data_files
    )
    etld_lib_config.remove_old_files(
        dir_path=etld_lib_config.was_distribution_dir,
        dir_search_glob=etld_lib_config.was_distribution_dir_file_search_blob
    )

    was_count_webapp_arg_dict = {
        'api_call_name': 'was_count_webapp',
        'was_extract_function_reference': was_04_extract_from_qualys.was_webapp_extract_count,
        'utc_datetime': utc_datetime, 'cred_dict': cred_dict,
        'dir_search_glob': etld_lib_config.was_extract_dir_file_search_blob_webapp_count
    }
    was_count_webapp_proc = multiprocessing.Process(
        target=get_was_count,
        name='was_count_webapp',
        args=(was_count_webapp_arg_dict,)
    )
    was_count_catalog_arg_dict = {
        'api_call_name': 'was_count_catalog',
        'was_extract_function_reference': was_04_extract_from_qualys.was_catalog_extract_count,
        'utc_datetime': utc_datetime, 'cred_dict': cred_dict,
        'last_id': etld_lib_config.was_catalog_start_greater_than_last_id,
        'dir_search_glob': etld_lib_config.was_extract_dir_file_search_blob_catalog_count
    }
    was_count_catalog_proc = multiprocessing.Process(
        target=get_was_count,
        name='was_count_catalog',
        args=(was_count_catalog_arg_dict,)
    )

    was_webapp_arg_dict = {
        'api_call_name': 'was_webapp',
        'was_extract_function_reference': was_04_extract_from_qualys.was_webapp_extract,
        'utc_datetime': utc_datetime, 'cred_dict': cred_dict, 'page_size': 300,
        'dir_search_glob': etld_lib_config.was_extract_dir_file_search_blob_webapp
    }
    was_webapp_proc = multiprocessing.Process(
        target=get_was_data,
        name='was_webapp',
        args=(was_webapp_arg_dict,)
    )

    was_catalog_arg_dict = {
        'api_call_name': 'was_catalog',
        'was_extract_function_reference': was_04_extract_from_qualys.was_catalog_extract,
        'utc_datetime': utc_datetime, 'cred_dict': cred_dict, 'page_size': 300,
        'last_id': etld_lib_config.was_catalog_start_greater_than_last_id,
        'dir_search_glob': etld_lib_config.was_extract_dir_file_search_blob_catalog
    }
    was_catalog_proc = multiprocessing.Process(
        target=get_was_data,
        name='was_catalog',
        args=(was_catalog_arg_dict,)
    )

    proc_list = [
        was_count_webapp_proc,
        was_count_catalog_proc,
        was_webapp_proc,
        was_catalog_proc
    ]

    for proc_item in proc_list:
        proc_item.daemon = True
        proc_item.start()

    time.sleep(15)
    proc_timeout = 36000  # 10 hour catch all
    for proc_item in proc_list:
        proc_item.join(timeout=proc_timeout)

    for proc_item in proc_list:
        if proc_item.exitcode == 0:
            etld_lib_functions.logger.info(f"{proc_item.name} exited successfully, exit code: {proc_item.exitcode}")
        else:
            etld_lib_functions.logger.error(f"{proc_item.name} exited after .join timeout of {proc_timeout} seconds "
                                            f"with error code: {proc_item.exitcode}")
            etld_lib_functions.logger.error(f"{proc_item.name} failed, please investigate why job ran so long and rerun.")
            for proc_item_to_kill in proc_list:
                proc_item_to_kill.kill()
            exit(1)

    end_was_03_extract_controller()


def main():
    try:
        extract_was_api_data()
    except Exception as e:
        etld_lib_functions.logger.error(f"ERROR: extract_was_api_data function failed.")
        etld_lib_functions.logger.error(f"ERROR: {e}")
        formatted_lines = traceback.format_exc().splitlines()
        etld_lib_functions.logger.error(f"ERROR: {formatted_lines}")


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='was_03_extract_controller')
    etld_lib_config.main()
    etld_lib_credentials.main()
    etld_lib_authentication_objects.main()
    main()



#!/usr/bin/env python3
import xmltodict
import json
import re
import time
import sqlite3
from multiprocessing import Process, Queue
from pathlib import Path
import gzip

from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions
# from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects
from qualys_etl.etld_lib import etld_lib_sqlite_tables
from qualys_etl.etld_lib import etld_lib_extract_transform_load
from qualys_etl.etld_lib import etld_lib_datetime
from qualys_etl.etld_lib import etld_lib_csv_distribution
from qualys_etl.etld_host_list_detection import host_list_detection_02_workflow_manager
from qualys_etl.etld_knowledgebase import knowledgebase_03_extract_controller
import codecs


def create_counter_objects(database_type='sqlite'):
    counter_obj_host_list_detection_hosts = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=10000,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"rows added to {database_type} "
                                    f"table {etld_lib_config.host_list_detection_hosts_table_name}")

    counter_obj_host_list_detection_qids = \
        etld_lib_functions.DisplayCounterToLog(
            display_counter_at=100000,
            logger_func=etld_lib_functions.logger.info,
            display_counter_log_message=f"rows added to {database_type} "
                                        f"table {etld_lib_config.host_list_detection_qids_table_name}")

    counter_obj_host_list_detection_hosts_without_qids = \
        etld_lib_functions.DisplayCounterToLog(
            display_counter_at=10000,
            logger_func=etld_lib_functions.logger.info,
            display_counter_log_message=f"hosts without detection rows added to {database_type} "
                                        f"table {etld_lib_config.host_list_detection_qids_table_name}")

    counter_obj_dict_new = {
        'counter_obj_host_list_detection_hosts': counter_obj_host_list_detection_hosts,
        'counter_obj_host_list_detection_qids': counter_obj_host_list_detection_qids,
        'counter_obj_host_list_detection_hosts_without_qids': counter_obj_host_list_detection_hosts_without_qids,
    }

    return counter_obj_dict_new


def drop_and_create_all_tables(sqlite_obj):
    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.host_list_table_name,
        csv_columns=etld_lib_config.host_list_csv_columns(),
        csv_column_types=etld_lib_config.host_list_csv_column_types(),
        key='ID')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.host_list_detection_q_knowledgebase_in_host_list_detection,
        csv_columns=etld_lib_config.kb_csv_columns(),
        csv_column_types=etld_lib_config.kb_csv_column_types(),
        key='QID')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.host_list_detection_hosts_table_name,
        csv_columns=etld_lib_config.host_list_detection_host_csv_columns(),
        csv_column_types=etld_lib_config.host_list_detection_host_csv_column_types(),
        key='ID')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.host_list_detection_qids_table_name,
        csv_columns=etld_lib_config.host_list_detection_qids_csv_columns(),
        csv_column_types=etld_lib_config.host_list_detection_qids_csv_column_types(),
    )

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.host_list_detection_status_table_name,
        csv_columns=etld_lib_config.status_table_csv_columns(),
        csv_column_types=etld_lib_config.status_table_csv_column_types(),
        key='STATUS_NAME')

    sqlite_obj.execute_statement(f"DROP VIEW IF EXISTS {etld_lib_config.host_list_detection_table_view_name}")
    # create_view_statement = \
    #     '''CREATE VIEW Q_Host_List_Detection as
    # select A.ID HL_ID, A.ASSET_ID HL_ASSET_ID, A.IP HL_IP, A.IPV6 HL_IPV6, A.TRACKING_METHOD HL_TRACKING_METHOD,
    # A.NETWORK_ID HL_NETWORK_ID, A.DNS HL_DNS, A.DNS_DATA HL_DNS_DATA, A.CLOUD_PROVIDER HL_CLOUD_PROVIDER,
    # A.CLOUD_SERVICE HL_CLOUD_SERVICE, A.CLOUD_RESOURCE_ID HL_CLOUD_RESOURCE_ID,
    # A.EC2_INSTANCE_ID HL_EC2_INSTANCE_ID, A.NETBIOS HL_NETBIOS, A.OS HL_OS, A.QG_HOSTID HL_QG_HOSTID,
    # A.TAGS HL_TAGS, A.METADATA HL_METADATA, A.CLOUD_PROVIDER_TAGS HL_CLOUD_PROVIDER_TAGS,
    # A.LAST_VULN_SCAN_DATETIME HL_LAST_VULN_SCAN_DATETIME, A.LAST_VM_SCANNED_DATE HL_LAST_VM_SCANNED_DATE,
    # A.LAST_VM_SCANNED_DURATION HL_LAST_VM_SCANNED_DURATION,
    # A.LAST_VM_AUTH_SCANNED_DATE HL_LAST_VM_AUTH_SCANNED_DATE,
    # A.LAST_VM_AUTH_SCANNED_DURATION HL_LAST_VM_AUTH_SCANNED_DURATION,
    # A.LAST_COMPLIANCE_SCAN_DATETIME HL_LAST_COMPLIANCE_SCAN_DATETIME, A.OWNER HL_OWNER,
    # A.COMMENTS HL_COMMENTS, A.USER_DEF HL_USER_DEF, A.ASSET_GROUP_IDS HL_ASSET_GROUP_IDS,
    # A.ASSET_RISK_SCORE HL_ASSET_RISK_SCORE, A.ASSET_CRITICALITY_SCORE HL_ASSET_CRITICALITY_SCORE,
    # A.ARS_FACTORS HL_ARS_FACTORS,
    # A.BATCH_DATE HL_BATCH_DATE, A.BATCH_NUMBER HL_BATCH_NUMBER, A.Row_Last_Updated HL_Row_Last_Updated,
    # B.ID HLDH_ID, B.ASSET_ID HLDH_ASSET_ID, B.IP HLDH_IP, B.IPV6 HLDH_IPV6, B.TRACKING_METHOD HLDH_TRACKING_METHOD,
    # B.NETWORK_ID HLDH_NETWORK_ID, B.OS HLDH_OS, B.OS_CPE HLDH_OS_CPE, B.DNS HLDH_DNS, B.DNS_DATA HLDH_DNS_DATA,
    # B.NETBIOS HLDH_NETBIOS, B.QG_HOSTID HLDH_QG_HOSTID, B.LAST_SCAN_DATETIME HLDH_LAST_SCAN_DATETIME,
    # B.LAST_VM_SCANNED_DATE HLDH_LAST_VM_SCANNED_DATE, B.LAST_VM_SCANNED_DURATION HLDH_LAST_VM_SCANNED_DURATION,
    # B.LAST_VM_AUTH_SCANNED_DATE HLDH_LAST_VM_AUTH_SCANNED_DATE,
    # B.LAST_VM_AUTH_SCANNED_DURATION HLDH_LAST_VM_AUTH_SCANNED_DURATION,
    # B.LAST_PC_SCANNED_DATE HLDH_LAST_PC_SCANNED_DATE, B.BATCH_DATE HLDH_BATCH_DATE,
    # B.BATCH_NUMBER HLDH_BATCH_NUMBER, B.Row_Last_Updated HLDH_Row_Last_Updated,
    # C.ID HLDQ_ID, C.ASSET_ID HLDQ_ASSET_ID, C.QID HLDQ_QID, C.TYPE HLDQ_TYPE, C.STATUS HLDQ_STATUS,
    # C.PORT HLDQ_PORT, C.PROTOCOL HLDQ_PROTOCOL, C.SEVERITY HLDQ_SEVERITY, C.FQDN HLDQ_FQDN, C.SSL HLDQ_SSL,
    # C.INSTANCE HLDQ_INSTANCE, C.LAST_PROCESSED_DATETIME HLDQ_LAST_PROCESSED_DATETIME,
    # C.FIRST_FOUND_DATETIME HLDQ_FIRST_FOUND_DATETIME, C.LAST_FOUND_DATETIME HLDQ_LAST_FOUND_DATETIME,
    # C.TIMES_FOUND HLDQ_TIMES_FOUND, C.LAST_TEST_DATETIME HLDQ_LAST_TEST_DATETIME,
    # C.LAST_UPDATE_DATETIME HLDQ_LAST_UPDATE_DATETIME, C.LAST_FIXED_DATETIME HLDQ_LAST_FIXED_DATETIME,
    # C.FIRST_REOPENED_DATETIME HLDQ_FIRST_REOPENED_DATETIME, C.LAST_REOPENED_DATETIME HLDQ_LAST_REOPENED_DATETIME,
    # C.TIMES_REOPENED HLDQ_TIMES_REOPENED, C.SERVICE HLDQ_SERVICE, C.IS_IGNORED HLDQ_IS_IGNORED,
    # C.IS_DISABLED HLDQ_IS_DISABLED, C.AFFECT_RUNNING_KERNEL HLDQ_AFFECT_RUNNING_KERNEL,
    # C.AFFECT_RUNNING_SERVICE HLDQ_AFFECT_RUNNING_SERVICE,
    # C.AFFECT_EXPLOITABLE_CONFIG HLDQ_AFFECT_EXPLOITABLE_CONFIG,
    # C.QDS HLDQ_QDS, C.QDS_FACTORS HLDQ_QDS_FACTORS,
    # C.RESULTS HLDQ_RESULTS,
    # C.BATCH_DATE HLDQ_BATCH_DATE,
    # C.BATCH_NUMBER HLDQ_BATCH_NUMBER,
    # C.Row_Last_Updated HLDQ_Row_Last_Updated
    # from Q_Host_List A
    # left outer join Q_Host_List_Detection_HOSTS B ON A.ID = B.ID
    # left outer join Q_Host_List_Detection_QIDS C ON A.ID = C.ID'''
    #
    # sqlite_obj.execute_statement(create_view_statement)
    # create_view_statement_display = create_view_statement.replace('\n', '').replace('  ', ' ')
    # etld_lib_functions.logger.info(f"VIEW CREATED: {create_view_statement_display}")
    sqlite_obj.commit_changes()


def attach_database_to_host_list_detection(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj, table_name, database_as_name, database_file):
    sqlite_obj.attach_database_to_connection(database_as_name=database_as_name,
                                             database_sqlite_file=database_file)
    table_name = f"{database_as_name}.{table_name}"
    table_columns = sqlite_obj.get_table_columns(table_name=table_name)
    etld_lib_functions.logger.info(f"Found Table {table_name} columns: {table_columns}")
    return table_name, table_columns


def detach_database_to_host_list_detection(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj, table_name, database_as_name, database_file):
    sqlite_obj.detach_database_to_connection(database_as_name=database_as_name,
                                             database_sqlite_file=database_file)


def get_batch_date_from_status_table(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj, status_table_name, table_name):
    batch_date = None
    try:
        etld_lib_functions.logger.info(f"Begin retrieving batch_date for {table_name}")
        rows = sqlite_obj.select_all_from_status_table_where_status_name_equals_table_name(
            status_table_name=status_table_name,
            table_name=table_name
        )
        for row in rows:
            status_detail_dict = json.loads(row['STATUS_DETAIL'])
            batch_date = status_detail_dict['BATCH_DATE']
            break  # Any row will do.
        etld_lib_functions.logger.info(f"End   retrieving batch_date for {table_name}")
    except Exception as e:
        etld_lib_functions.logger.error(f"Error retrieving batch_date for {table_name}")
        etld_lib_functions.logger.error(f"Exception is: {e}")
        exit(1)
    return batch_date


# def copy_database_table(
#         sqlite_obj: etld_lib_sqlite_tables.SqliteObj, table_name, new_table_name, where_clause=""):
#     try:
#         etld_lib_functions.logger.info(f"Begin creating {new_table_name} from {table_name}")
#         #        etl_workflow_sqlite_obj.cursor.execute(f"DROP TABLE IF EXISTS {new_table_name}")
#         #        etl_workflow_sqlite_obj.cursor.execute(f"CREATE TABLE {new_table_name} AS SELECT * FROM {table_name} where 1=0")
#         # TODO CHECK ME
#         sqlite_obj.cursor.execute(f"INSERT INTO {new_table_name} SELECT * FROM {table_name} {where_clause}")
#         etld_lib_functions.logger.info(f"End   creating {new_table_name} from {table_name}")
#     except Exception as e:
#         etld_lib_functions.logger.error(f"Error creating table {new_table_name} from {table_name}")
#         etld_lib_functions.logger.error(f"Exception is: {e}")
#         exit(1)


def copy_database_table(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj, table_name, new_table_name, where_clause="", select_clause="*"):
    max_attempts = 15  # Maximum number of attempts
    attempt = 0
    sleep_time = 15
    success_flag = False

    while attempt < max_attempts:
        try:
            etld_lib_functions.logger.info(f"Attempt {attempt + 1}: Begin creating {new_table_name} from {table_name}")

            etld_lib_functions.logger.info(
                f"INSERT INTO {new_table_name} SELECT {select_clause} FROM {table_name} {where_clause}")
            sqlite_obj.cursor.execute(
                f"INSERT INTO {new_table_name} SELECT {select_clause} FROM {table_name} {where_clause}")
            time.sleep(sleep_time)
            etld_lib_functions.logger.info(f"Attempt {attempt + 1}: End creating {new_table_name} from {table_name}")
            success_flag = True
            break  # Success, exit the loop
        except sqlite3.OperationalError as oe:
            time.sleep(sleep_time)
            etld_lib_functions.logger.warning(f"Operational error on attempt {attempt + 1}: {oe}")
        except sqlite3.IntegrityError as ie:
            time.sleep(sleep_time)
            etld_lib_functions.logger.warning(f"Integrity error on attempt {attempt + 1}: {ie}")
            break
        except sqlite3.DatabaseError as de:
            time.sleep(sleep_time)
            etld_lib_functions.logger.warning(f"Database error on attempt {attempt + 1}: {de}")
        except sqlite3.Error as se:
            time.sleep(sleep_time)
            etld_lib_functions.logger.warning(f"sqlite3.Error on attempt {attempt + 1}: {se}")
        except IOError as ioe:
            time.sleep(sleep_time)
            etld_lib_functions.logger.warning(f"IO error on attempt {attempt + 1}: {ioe}")
        except Exception as e:
            time.sleep(sleep_time)
            etld_lib_functions.logger.warning(f"Unexpected error on attempt {attempt + 1}: {e}")

        attempt += 1
        if attempt < max_attempts:
            time.sleep(sleep_time)
            etld_lib_functions.logger.info(f"Retrying in {sleep_time} seconds...")
            time.sleep(sleep_time)  # Wait for 30 seconds before retrying

    if attempt == max_attempts:
        time.sleep(5)
        etld_lib_functions.logger.error(
            f"Failed to create table {new_table_name} from {table_name} after {max_attempts} attempts.")
        exit(1)

    if not success_flag:
        time.sleep(5)  # Wait for 30 seconds before retrying
        etld_lib_functions.logger.error(
            f"Success Flag is false; Failed to create table {new_table_name} from {table_name}")
        exit(1)


# def create_kb_and_host_list_tables_in_host_list_detection_database(etl_workflow_sqlite_obj: etld_lib_sqlite_tables.SqliteObj):
#     etld_lib_functions.logger.info(f"Attaching database: {etld_lib_config.host_list_sqlite_file}")
#     host_list_table_name, host_list_table_columns = \
#         attach_database_to_host_list_detection(etl_workflow_sqlite_obj=etl_workflow_sqlite_obj,
#                                                table_name="Q_Host_List",
#                                                database_as_name="H1",
#                                                database_file=etld_lib_config.host_list_sqlite_file)
#
#     # Update knowledgebase after extracting detections in case of edge case qid created during run.
#     host_list_detection_02_workflow_manager.get_knowledgebase_controller(
#        kb_last_modified_after=etld_lib_datetime.get_utc_date_minus_days(days=7))
#
#     etld_lib_functions.logger.info(f"Attaching database: {etld_lib_config.kb_table_name}")
#     kb_table_name, kb_table_columns = \
#         attach_database_to_host_list_detection(etl_workflow_sqlite_obj=etl_workflow_sqlite_obj,
#                                                table_name="Q_KnowledgeBase",
#                                                database_as_name="K1",
#                                                database_file=etld_lib_config.kb_sqlite_file)
#
#     etld_lib_functions.logger.info(f"Copying table: {host_list_table_name}")
#     copy_database_table(etl_workflow_sqlite_obj=etl_workflow_sqlite_obj, table_name=host_list_table_name, new_table_name="Q_Host_List")
#
#     etld_lib_functions.logger.info(f"Copying table: {kb_table_name}")
#     where_clause = f"where {kb_table_name}.QID in (select distinct(QID) from Q_Host_List_Detection_QIDS)"
#     copy_database_table(etl_workflow_sqlite_obj=etl_workflow_sqlite_obj, table_name=kb_table_name,
#                         new_table_name="Q_KnowledgeBase_In_Host_List_Detection", where_clause=where_clause)
#     etld_lib_functions.logger.info(f"Commit copy of table: {host_list_table_name}, {kb_table_name}")
#     etl_workflow_sqlite_obj.commit_changes()


def update_host_list_in_host_list_detection_database(sqlite_obj: etld_lib_sqlite_tables.SqliteObj):
    etld_lib_functions.logger.info(f"Attaching database: {etld_lib_config.host_list_sqlite_file}")
    host_list_table_name, host_list_table_columns = \
        attach_database_to_host_list_detection(sqlite_obj=sqlite_obj,
                                               table_name="Q_Host_List",
                                               database_as_name="H1",
                                               database_file=etld_lib_config.host_list_sqlite_file)

    etld_lib_functions.logger.info(f"Copying table: {host_list_table_name}")
    copy_database_table(sqlite_obj=sqlite_obj, table_name=host_list_table_name, new_table_name="Q_Host_List",
                        select_clause="*")
    etld_lib_functions.logger.info(f"Commit copy of table: {host_list_table_name}")
    sqlite_obj.commit_changes()

    etld_lib_functions.logger.info(f"Detaching database: {etld_lib_config.host_list_sqlite_file}")
    detach_database_to_host_list_detection(sqlite_obj=sqlite_obj,
                                           table_name="Q_Host_List",
                                           database_as_name="H1",
                                           database_file=etld_lib_config.host_list_sqlite_file)
    sqlite_obj.commit_changes()


def refresh_knowledgebase():
    kb_last_modified_after = etld_lib_datetime.get_utc_date_minus_days(days=30)
    knowledgebase_03_extract_controller.knowledgebase_extract_controller(
        kb_last_modified_after=kb_last_modified_after,
        lock_file_required=True
    )


def update_knowledgebase_in_host_list_detection_database(refresh_knowledgebase_flag=True):
    if refresh_knowledgebase_flag:
        refresh_knowledgebase()

    sqlite_obj = etld_lib_sqlite_tables.SqliteObj(
        sqlite_file=etld_lib_config.host_list_detection_sqlite_file)

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.host_list_detection_q_knowledgebase_in_host_list_detection,
        csv_columns=etld_lib_config.kb_csv_columns(),
        csv_column_types=etld_lib_config.kb_csv_column_types(),
        key='QID')

    etld_lib_functions.logger.info(f"Attaching database: {etld_lib_config.kb_table_name}")
    kb_table_name, kb_table_columns = \
        attach_database_to_host_list_detection(sqlite_obj=sqlite_obj,
                                               table_name="Q_KnowledgeBase",
                                               database_as_name="K1",
                                               database_file=etld_lib_config.kb_sqlite_file)

    # etld_lib_functions.logger.info(f"Copying table: {kb_table_name}")
    # where_clause = f"where {kb_table_name}.QID in (select distinct(QID) from Q_Host_List_Detection_QIDS)"
    # copy_database_table(sqlite_obj=sqlite_obj, table_name=kb_table_name,
    #                     new_table_name="Q_KnowledgeBase_In_Host_List_Detection", where_clause=where_clause)
    #
    # etld_lib_functions.logger.info(f"Commit copy of table:  {kb_table_name}")
    # sqlite_obj.commit_changes()
    sqlite_obj.create_q_knowledgebase_in_host_list_detection()

    etld_lib_functions.logger.info(f"Detaching database: {etld_lib_config.kb_table_name}")
    detach_database_to_host_list_detection(sqlite_obj=sqlite_obj,
                                           table_name="Q_KnowledgeBase",
                                           database_as_name="K1",
                                           database_file=etld_lib_config.kb_sqlite_file)
    sqlite_obj.commit_changes()
    sqlite_obj.close_connection()


def update_final_status_in_host_list_detection_database():
    sqlite_obj = etld_lib_sqlite_tables.SqliteObj(
        sqlite_file=etld_lib_config.host_list_detection_sqlite_file)
    batch_date = get_batch_date_from_status_table(
        sqlite_obj=sqlite_obj,
        status_table_name=etld_lib_config.host_list_detection_status_table_name,
        table_name=etld_lib_config.host_list_detection_hosts_table_name
    )
    sqlite_obj.update_status_table(
        batch_date=batch_date,
        batch_number="0",
        total_rows_added_to_database=0,
        status_table_name=etld_lib_config.host_list_detection_status_table_name,
        status_table_columns=etld_lib_config.status_table_csv_columns(),
        status_table_column_types=etld_lib_config.status_table_csv_column_types(),
        status_name_column='ALL_TABLES_LOADED_SUCCESSFULLY',
        status_column='final')
    sqlite_obj.commit_changes()
    sqlite_obj.close_connection()


def insert_into_sqlite_host_list_detection_host_and_associated_qids(
        host_list_detection_document: dict,
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
        counter_obj: dict,
        csv_distribution_dict: dict,
        batch_number: str = '000001',
        batch_date: str = '1970-01-01 00:00:00'
):
    # def prepare_database_field(field_data, field_name_tmp) -> dict:
    #     if field_data is None:
    #         field_data = ""
    #     elif 'DATE' in field_name_tmp:
    #         field_data = field_data.replace("T", " ").replace("Z", "")
    #         field_data = re.sub("\\..*$", "", field_data)
    #     elif isinstance(field_data, int):
    #         field_data = str(field_data)
    #     elif not isinstance(field_data, str):
    #         field_data = json.dumps(field_data)
    #
    #     return field_data

    # def prepare_database_row(item: dict, database_columns: list) -> list:
    #     row_in_sqlite_form = []
    #     for field_name in database_columns:
    #         if field_name in item.keys():
    #             item[field_name] = \
    #                 prepare_database_field(item[field_name], field_name)
    #             row_in_sqlite_form.append(item[field_name])
    #         else:
    #             row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field
    #     return row_in_sqlite_form

    def insert_host_rows_into_database_and_csv(host_document):
        # host_row_in_sqlite_form: list = \
        #     prepare_database_row(
        #         item=host_document,
        #         database_columns=etld_lib_config.host_list_detection_host_csv_columns())

        # host_row_in_sqlite_form: list = \
        #     sqlite_obj.prepare_database_row(
        #         item=host_document,
        #         database_columns=etld_lib_config.host_list_detection_host_csv_columns(),
        #         database_column_types=etld_lib_config.host_list_detection_host_csv_column_types()
        #     )

        host_row_in_sqlite_form: list = \
            sqlite_obj.prepare_database_row_vmpc(
                item_dict=host_document,
                csv_columns=etld_lib_config.host_list_detection_host_csv_columns(),
                csv_column_types=etld_lib_config.host_list_detection_host_csv_column_types(),
                batch_date=batch_date,
                batch_number=batch_number
            )

        result = sqlite_obj.insert_or_replace_row_pristine(
            table_name=etld_lib_config.host_list_detection_hosts_table_name,
            row=host_row_in_sqlite_form)

        if result is True:
            counter_obj['counter_obj_host_list_detection_hosts'].update_counter_and_display_to_log()
            if etld_lib_config.host_list_detection_distribution_csv_flag:
                # CSV - WRITE ROW
                # Write one row to Q_Host_List_Detection_HOSTS CSV File.
                #
                csv_obj: etld_lib_csv_distribution.CsvDistribution = \
                    csv_distribution_dict['q_host_list_detection_hosts_csv_obj']
                csv_obj.write_csv_row_to_file_handle(row=host_row_in_sqlite_form)

        else:
            etld_lib_functions.logger.error("Error inserting detection")
            row = re.sub('\n', '', '|'.join(host_row_in_sqlite_form))
            json_str = json.dumps(host_document)
            json_str = re.sub('\n', '', json_str)
            etld_lib_functions.logger.error(f"ROW: {row}")
            etld_lib_functions.logger.error(f"JSON: {json_str}")
            exit(1)

    def insert_qids_rows_into_database_and_csv(detection_list: list):
        if isinstance(detection_list, dict):
            detection_list = [detection_list]

        for one_detection_dict in detection_list:
            if isinstance(one_detection_dict, dict):
                one_detection_dict['ID'] = host_list_detection_document['ID']
                if 'ASSET_ID' in host_list_detection_document.keys():
                    one_detection_dict['ASSET_ID'] = host_list_detection_document['ASSET_ID']
                # one_detection_dict['BATCH_DATE'] = batch_date
                # one_detection_dict['BATCH_NUMBER'] = batch_number

                qids_row_in_sqlite_form = \
                    sqlite_obj.prepare_database_row_vmpc(
                        item_dict=one_detection_dict,
                        csv_columns=etld_lib_config.host_list_detection_qids_csv_columns(),
                        csv_column_types=etld_lib_config.host_list_detection_qids_csv_column_types(),
                        batch_date=batch_date,
                        batch_number=batch_number
                    )

                result = sqlite_obj.insert_or_replace_row_pristine(
                    table_name=etld_lib_config.host_list_detection_qids_table_name,
                    row=qids_row_in_sqlite_form)

                if result is True:
                    counter_obj['counter_obj_host_list_detection_qids'].update_counter_and_display_to_log()
                    if etld_lib_config.host_list_detection_distribution_csv_flag:
                        #
                        # Write one row to Q_Host_List_Detection_QIDS CSV File.
                        #
                        csv_obj: etld_lib_csv_distribution.CsvDistribution = \
                            csv_distribution_dict['q_host_list_detection_qids_csv_obj']
                        csv_obj.write_csv_row_to_file_handle(qids_row_in_sqlite_form)
                else:
                    etld_lib_functions.logger.error("Error inserting detection")
                    row = re.sub('\n', '', '|'.join(qids_row_in_sqlite_form))
                    json_str = json.dumps(one_detection_dict)
                    json_str = re.sub('\n', '', json_str)
                    etld_lib_functions.logger.error(f"ROW: {row}")
                    etld_lib_functions.logger.error(f"JSON: {json_str}")
                    exit(1)

    def qids_found():
        qids_found_flag = False
        if 'DETECTION_LIST' in host_list_detection_document:
            if 'DETECTION' in host_list_detection_document['DETECTION_LIST']:
                qids_found_flag = True
            else:
                json_message = \
                    {'ID': host_list_detection_document['ID'], 'BATCH_NUMBER': batch_number, 'BATCH_DATE': batch_date}
                etld_lib_functions.logger.warning(f"No DETECTION for: {json_message}")
        else:
            json_message = \
                {'ID': host_list_detection_document['ID'], 'BATCH_NUMBER': batch_number, 'BATCH_DATE': batch_date}
            etld_lib_functions.logger.warning(f"No DETECTION_LIST for: {json_message}")
        return qids_found_flag

    #
    # Insert Host and QIDS into targets database and csv
    #
    # host_list_detection_document['BATCH_DATE'] = batch_date
    # host_list_detection_document['BATCH_NUMBER'] = batch_number
    insert_host_rows_into_database_and_csv(host_list_detection_document)
    if qids_found():
        insert_qids_rows_into_database_and_csv(host_list_detection_document['DETECTION_LIST']['DETECTION'])


# def q_hosts_csv_info_setup(batch_date, batch_number) -> dict:
#     q_host_list_detection_hosts_target_csv_file_name, \
#     table_name_utc_run_datetime,  \
#     table_name_batch_number,  \
#     q_host_list_detection_hosts_target_tmp_csv_file_name = \
#         etld_lib_csv_distribution.get_target_csv_file_name(
#             batch_date, batch_number,
#             etld_lib_config.host_list_detection_hosts_table_name,
#             csv_file_open_method=etld_lib_config.host_list_detection_open_file_compression_method
#         )
#     # FINAL CSV FILENAME Target
#     q_host_list_detection_hosts_target_csv_file_path = \
#         Path(etld_lib_config.host_list_detection_distribution_dir,
#              q_host_list_detection_hosts_target_csv_file_name)
#     # TEMP CSV FILENAME WHILE BUILDING FILE
#     q_host_list_detection_hosts_target_tmp_csv_file_path = \
#         Path(etld_lib_config.host_list_detection_distribution_dir,
#              q_host_list_detection_hosts_target_tmp_csv_file_name
#              )
#
#     q_hosts_csv_info = {}
#     q_hosts_csv_info['target_csv_file_name'] = q_host_list_detection_hosts_target_csv_file_name
#     q_hosts_csv_info['table_name_utc_run_datetime'] = table_name_utc_run_datetime
#     q_hosts_csv_info['table_name_batch_number'] = table_name_batch_number
#     q_hosts_csv_info['target_tmp_csv_file_name'] = q_host_list_detection_hosts_target_tmp_csv_file_name
#     q_hosts_csv_info['target_csv_file_path'] = q_host_list_detection_hosts_target_csv_file_path
#     q_hosts_csv_info['target_tmp_csv_file_path'] = q_host_list_detection_hosts_target_tmp_csv_file_path
#     q_hosts_csv_info['distribution_dir'] = etld_lib_config.host_list_detection_distribution_dir
#     q_hosts_csv_info['compression_method'] = etld_lib_config.host_list_detection_open_file_compression_method
#     #
#     # Create Q_Host_List_Detection_HOSTS CSV OBJECT
#     #
#     q_hosts_csv_info['csv_obj'] = \
#         etld_lib_csv_distribution.CsvDistribution(
#             target_csv_path=q_hosts_csv_info['target_tmp_csv_file_path'],
#             target_csv_data_directory=Path(q_hosts_csv_info['distribution_dir']))
#
#     return q_hosts_csv_info


# def q_qids_csv_info_setup(batch_date, batch_number) -> dict:
#     q_host_list_detection_qids_target_csv_file_name, \
#     table_name_utc_run_datetime, \
#     table_name_batch_number, \
#     q_host_list_detection_qids_target_tmp_csv_file_name = \
#         etld_lib_csv_distribution.get_target_csv_file_name(
#             batch_date, batch_number,
#             etld_lib_config.host_list_detection_qids_table_name,
#             csv_file_open_method=etld_lib_config.host_list_detection_open_file_compression_method
#         )
#
#     # FINAL CSV FILENAME Target
#     q_host_list_detection_qids_target_csv_file_path = \
#         Path(etld_lib_config.host_list_detection_distribution_dir,
#              q_host_list_detection_qids_target_csv_file_name)
#     # TEMP CSV FILENAME WHILE BUILDING FILE
#     q_host_list_detection_qids_target_tmp_csv_file_path = \
#         Path(etld_lib_config.host_list_detection_distribution_dir,
#              q_host_list_detection_qids_target_tmp_csv_file_name)
#
#     q_qids_csv_info = {}
#     q_qids_csv_info['target_csv_file_name'] = q_host_list_detection_qids_target_csv_file_name
#     q_qids_csv_info['table_name_utc_run_datetime'] = table_name_utc_run_datetime
#     q_qids_csv_info['table_name_batch_number'] = table_name_batch_number
#     q_qids_csv_info['tmp_csv_file_name'] = q_host_list_detection_qids_target_tmp_csv_file_name
#     q_qids_csv_info['target_csv_file_path'] = q_host_list_detection_qids_target_csv_file_path
#     q_qids_csv_info['target_tmp_csv_file_path'] = q_host_list_detection_qids_target_tmp_csv_file_path
#     q_qids_csv_info['distribution_dir'] = etld_lib_config.host_list_detection_distribution_dir
#     q_qids_csv_info['compression_method'] = etld_lib_config.host_list_detection_open_file_compression_method
#
#     #
#     # Create Q_Host_List_Detection_QIDS CSV OBJECT
#     #
#
#     q_qids_csv_info['csv_obj'] = \
#         etld_lib_csv_distribution.CsvDistribution(
#             target_csv_path=q_qids_csv_info['target_tmp_csv_file_path'],
#             target_csv_data_directory=Path(q_qids_csv_info['distribution_dir']))
#
#     return q_qids_csv_info


def insert_xml_file_into_sqlite(xml_file, sqlite_obj: etld_lib_sqlite_tables.SqliteObj, counter_obj):
    def callback_to_insert_host_into_sqlite_and_csv(element_names: tuple, document_item: dict):
        if len(element_names) > 2 and "HOST" != element_names[3][0]:
            return True
        else:
            batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(xml_file)
            batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(xml_file)
            try:
                # CSV - CSV_OBJ
                insert_into_sqlite_host_list_detection_host_and_associated_qids(
                    host_list_detection_document=document_item,
                    sqlite_obj=sqlite_obj,
                    batch_date=batch_date,
                    batch_number=batch_number,
                    counter_obj=counter_obj,
                    csv_distribution_dict=csv_distribution_dict
                )

            except Exception as e:
                etld_lib_functions.logger.error(f"Exception: {e}")
                etld_lib_functions.logger.error(
                    f"Issue inserting xml file into sqlite: {document_item}, counter={counter_obj}")
                exit(1)
            return True

    # READ XML FILE
    # with etld_lib_config.host_list_detection_open_file_compression_method(
    #         str(xml_file), "rt", encoding='utf-8') as xml_file_fd:
    if etld_lib_config.xmltodict_parse_using_codec_to_replace_utf8_error is True:
        xml_file_args = (str(xml_file), "rb")
        xml_file_kwargs = {}
    else:
        xml_file_args = (str(xml_file), "rt")
        xml_file_kwargs = {"encoding": "utf-8"}
    # 2024-01-26 with etld_lib_config.host_list_detection_open_file_compression_method(str(xml_file), "rb") as xml_file_fd:
    with etld_lib_config.host_list_detection_open_file_compression_method(*xml_file_args,
                                                                          **xml_file_kwargs) as xml_file_fd:

        batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(xml_file)
        batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(xml_file)

        csv_distribution_dict = {}
        if etld_lib_config.host_list_detection_distribution_csv_flag:
            # CSV - INFO
            # Create csv_info with all file and object locations for csv creation
            #
            q_hosts_csv_info = etld_lib_csv_distribution.get_csv_info_dict(
                batch_date=batch_date,
                batch_number=batch_number,
                table_name=etld_lib_config.host_list_detection_hosts_table_name,
                csv_file_open_method=etld_lib_config.host_list_detection_open_file_compression_method,
                distribution_dir=etld_lib_config.host_list_detection_distribution_dir
            )
            q_qids_csv_info = etld_lib_csv_distribution.get_csv_info_dict(
                batch_date=batch_date,
                batch_number=batch_number,
                table_name=etld_lib_config.host_list_detection_qids_table_name,
                csv_file_open_method=etld_lib_config.host_list_detection_open_file_compression_method,
                distribution_dir=etld_lib_config.host_list_detection_distribution_dir
            )
            # CSV - OPEN FILES FOR WRITING
            # BEGIN WRITING DATABASE AND CSV FILES
            #

            with q_hosts_csv_info['compression_method'](q_hosts_csv_info['target_tmp_csv_file_path'], 'wt',
                                                        newline='') as \
                    q_hosts_target_tmp_csv_file_handle:
                with q_qids_csv_info['compression_method'](q_qids_csv_info['target_tmp_csv_file_path'], 'wt',
                                                           newline='') as \
                        q_qids_target_tmp_csv_file_handle:
                    # CSV - SET FILE HANDLE FOR WRITE ROW
                    # Set csv writer file handles in csv_obj for reuse in callback_to_insert_host_into_sqlite_and_csv
                    q_hosts_csv_info['csv_obj'].set_csv_writer_with_file_handle(q_hosts_target_tmp_csv_file_handle)
                    q_qids_csv_info['csv_obj'].set_csv_writer_with_file_handle(q_qids_target_tmp_csv_file_handle)

                    csv_distribution_dict['q_host_list_detection_hosts_csv_obj'] = q_hosts_csv_info['csv_obj']
                    csv_distribution_dict['q_host_list_detection_qids_csv_obj'] = q_qids_csv_info['csv_obj']
                    # Insert into database and csv
                    # 2024-01-26 - updated for codec errors.
                    if etld_lib_config.xmltodict_parse_using_codec_to_replace_utf8_error is False:
                        xmltodict.parse(xml_file_fd.read(),
                                        item_depth=4,
                                        item_callback=callback_to_insert_host_into_sqlite_and_csv
                                        )
                    else:
                        xmltodict.parse(codecs.decode(xml_file_fd.read(), encoding='utf-8', errors='replace'),
                                        item_depth=4,
                                        item_callback=callback_to_insert_host_into_sqlite_and_csv
                                        )

            # Rename temp csv files to final naming .csv.gz.tmp to .csv.gz
            Path(q_hosts_csv_info['target_tmp_csv_file_path']).rename(q_hosts_csv_info['target_csv_file_path'])
            Path(q_qids_csv_info['target_tmp_csv_file_path']).rename(q_qids_csv_info['target_csv_file_path'])
        else:
            if etld_lib_config.xmltodict_parse_using_codec_to_replace_utf8_error is False:
                xmltodict.parse(xml_file_fd.read(),
                                item_depth=4,
                                item_callback=callback_to_insert_host_into_sqlite_and_csv
                                )
            else:
                xmltodict.parse(codecs.decode(xml_file_fd.read(), encoding='utf-8', errors='replace'),
                                item_depth=4,
                                item_callback=callback_to_insert_host_into_sqlite_and_csv
                                )

        sqlite_obj.commit_changes()
        csv_distribution_dict = None


def load_one_file_into_sqlite(sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
                              file_path: Path,
                              counter_obj: dict,
                              from_queue_or_directory="Queue",
                              batch_number: int = 0,
                              batch_name: str = "batch_000000",
                              batch_date: str = "2022-01-01 00:00:00",
                              ):
    etld_lib_functions.logger.info(f"Received batch file from {from_queue_or_directory}: {batch_name}")
    sqlite_obj.update_status_table(
        batch_date=batch_date, batch_number=batch_number,
        total_rows_added_to_database=counter_obj['counter_obj_host_list_detection_hosts'].get_counter(),
        status_table_name=etld_lib_config.host_list_detection_status_table_name,
        status_table_columns=etld_lib_config.status_table_csv_columns(),
        status_table_column_types=etld_lib_config.status_table_csv_column_types(),
        status_name_column=etld_lib_config.host_list_detection_hosts_table_name,
        status_column='begin')
    try:
        insert_xml_file_into_sqlite(file_path, sqlite_obj, counter_obj)
    except Exception as e:
        etld_lib_functions.logger.error(f"Exception: {e}")
        etld_lib_functions.logger.error(f"Issue inserting xml file into sqlite: {file_path}, counter={counter_obj}")
        exit(1)

    sqlite_obj.update_status_table(
        batch_date=batch_date, batch_number=batch_number,
        total_rows_added_to_database=counter_obj['counter_obj_host_list_detection_hosts'].get_counter(),
        status_table_name=etld_lib_config.host_list_detection_status_table_name,
        status_table_columns=etld_lib_config.status_table_csv_columns(),
        status_table_column_types=etld_lib_config.status_table_csv_column_types(),
        status_name_column=etld_lib_config.host_list_detection_hosts_table_name,
        status_column='end')
    etld_lib_functions.logger.info(f"Committed batch file to Database: {batch_name}")


def spawn_multiprocessing_queue_to_transform_and_load_xml_files_into_sqlite():
    queue_of_file_paths_to_load_to_sqlite = Queue()
    queue_process_to_load_to_sqlite = \
        Process(
            target=host_list_detection_transform_and_load_all_xml_files_into_sqlite,
            args=(queue_of_file_paths_to_load_to_sqlite, True),
            name="load_all_xml_files_into_sqlite")
    queue_process_to_load_to_sqlite.daemon = True
    queue_process_to_load_to_sqlite.start()

    queue_of_file_paths_to_load_to_sqlite.put("BEGIN")
    etld_lib_functions.logger.info(f"Queue of files process id: {queue_process_to_load_to_sqlite.pid} ")

    return queue_process_to_load_to_sqlite, queue_of_file_paths_to_load_to_sqlite


def load_files_into_sqlite_via_multiprocessing_queue(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
        queue_of_file_paths,
        counter_obj=None,
):
    def get_next_file_in_queue(bookend, queue_file_path):
        time.sleep(2)
        queue_data = queue_file_path.get()
        if queue_data == bookend:
            etld_lib_functions.logger.info(f"Found {bookend} of Queue.")
            queue_data = bookend
        return queue_data

    file_path = get_next_file_in_queue('BEGIN', queue_of_file_paths)
    batch_number = ""
    batch_date = ""
    if file_path == 'BEGIN':
        while True:
            file_path = get_next_file_in_queue('END', queue_of_file_paths)
            if file_path == 'END':
                sqlite_obj.update_status_table(
                    batch_date=batch_date, batch_number=batch_number,
                    total_rows_added_to_database=counter_obj['counter_obj_host_list_detection_hosts'].get_counter(),
                    status_table_name=etld_lib_config.host_list_detection_status_table_name,
                    status_table_columns=etld_lib_config.status_table_csv_columns(),
                    status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                    status_name_column=etld_lib_config.host_list_detection_hosts_table_name,
                    status_column='final')

                sqlite_obj.update_status_table(
                    batch_date=batch_date, batch_number=batch_number,
                    total_rows_added_to_database=counter_obj['counter_obj_host_list_detection_qids'].get_counter(),
                    status_table_name=etld_lib_config.host_list_detection_status_table_name,
                    status_table_columns=etld_lib_config.status_table_csv_columns(),
                    status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                    status_name_column=etld_lib_config.host_list_detection_qids_table_name,
                    status_column='final')

                # etl_workflow_sqlite_obj.update_status_table(
                #     batch_date=batch_date,
                #     batch_number=0,
                #     total_rows_added_to_database=0,
                #     etl_workflow_status_table_name=etld_lib_config.host_list_detection_status_table_name,
                #     status_table_columns=etld_lib_config.status_table_csv_columns(),
                #     status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                #     status_name_column='ALL_TABLES_LOADED_SUCCESSFULLY',
                #     status_column='final')
                break  # SUCCESSFUL END

            batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(file_path)
            batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(file_path)
            batch_name = etld_lib_extract_transform_load.get_batch_name_from_filename(file_path)
            load_one_file_into_sqlite(sqlite_obj=sqlite_obj,
                                      file_path=file_path,
                                      counter_obj=counter_obj,
                                      from_queue_or_directory="Queue",
                                      batch_number=batch_number,
                                      batch_name=batch_name,
                                      batch_date=batch_date,
                                      )
            sqlite_obj.commit_changes()
    else:
        etld_lib_functions.logger.error(f"Invalid begin of Queue, {file_path}.  Please restart.")
        exit(1)


def load_files_into_sqlite_via_directory_listing(sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
                                                 extract_dir,
                                                 extract_dir_file_search_blob,
                                                 counter_obj,
                                                 ):
    xml_file_list = []
    for file_name in sorted(Path(extract_dir).glob(extract_dir_file_search_blob)):
        if str(file_name).endswith('.xml') or str(file_name).endswith('.xml.gz'):
            xml_file_list.append(file_name)

    batch_number = ""
    batch_date = ""

    for file_path in xml_file_list:
        batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(file_path)
        batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(file_path)
        batch_name = etld_lib_extract_transform_load.get_batch_name_from_filename(file_path)
        load_one_file_into_sqlite(sqlite_obj=sqlite_obj,
                                  file_path=file_path,
                                  counter_obj=counter_obj,
                                  from_queue_or_directory="directory",
                                  batch_number=batch_number,
                                  batch_name=batch_name,
                                  batch_date=batch_date,
                                  )
        sqlite_obj.commit_changes()

    sqlite_obj.update_status_table(
        batch_date=batch_date, batch_number=batch_number,
        total_rows_added_to_database=counter_obj['counter_obj_host_list_detection_hosts'].get_counter(),
        status_table_name=etld_lib_config.host_list_detection_status_table_name,
        status_table_columns=etld_lib_config.status_table_csv_columns(),
        status_table_column_types=etld_lib_config.status_table_csv_column_types(),
        status_name_column=etld_lib_config.host_list_detection_hosts_table_name,
        status_column='final')

    sqlite_obj.update_status_table(
        batch_date=batch_date, batch_number=batch_number,
        total_rows_added_to_database=counter_obj['counter_obj_host_list_detection_qids'].get_counter(),
        status_table_name=etld_lib_config.host_list_detection_status_table_name,
        status_table_columns=etld_lib_config.status_table_csv_columns(),
        status_table_column_types=etld_lib_config.status_table_csv_column_types(),
        status_name_column=etld_lib_config.host_list_detection_qids_table_name,
        status_column='final')

    sqlite_obj.commit_changes()

    # etl_workflow_sqlite_obj.update_status_table(
    #     batch_date=batch_date,
    #     batch_number=0,
    #     total_rows_added_to_database=0,
    #     etl_workflow_status_table_name=etld_lib_config.host_list_detection_status_table_name,
    #     status_table_columns=etld_lib_config.status_table_csv_columns(),
    #     status_table_column_types=etld_lib_config.status_table_csv_column_types(),
    #     status_name_column='ALL_TABLES_LOADED_SUCCESSFULLY',
    #     status_column='final')


def host_list_detection_transform_and_load_all_xml_files_into_sqlite(
        queue_of_file_paths: Queue = Queue(), multiprocessing_flag=False):
    begin_host_list_detection_05_transform_load()
    xml_file_path = ""
    counter_obj_dict = create_counter_objects()
    try:
        host_list_detection_sqlite_obj = etld_lib_sqlite_tables.SqliteObj(
            sqlite_file=etld_lib_config.host_list_detection_sqlite_file)
        drop_and_create_all_tables(
            sqlite_obj=host_list_detection_sqlite_obj)
        #
        update_host_list_in_host_list_detection_database(sqlite_obj=host_list_detection_sqlite_obj)

        if multiprocessing_flag is True:
            load_files_into_sqlite_via_multiprocessing_queue(
                sqlite_obj=host_list_detection_sqlite_obj,
                queue_of_file_paths=queue_of_file_paths,
                counter_obj=counter_obj_dict
            )
        else:
            load_files_into_sqlite_via_directory_listing(
                sqlite_obj=host_list_detection_sqlite_obj,
                counter_obj=counter_obj_dict,
                extract_dir=etld_lib_config.host_list_detection_extract_dir,
                extract_dir_file_search_blob=etld_lib_config.host_list_detection_extract_dir_file_search_blob,
            )

        for counter_obj_key in counter_obj_dict.keys():
            counter_obj_dict[counter_obj_key].display_final_counter_to_log()

        # TODO update final batch number as multiprocessing results in out of order last batch
        # Move to end of program update_final_status_in_host_list_detection_database(sqlite_obj=host_list_detection_sqlite_obj)
        host_list_detection_sqlite_obj.close_connection()
        end_host_list_detection_05_transform_load()

    except Exception as e:
        etld_lib_functions.logger.error(f"Exception: {e}")
        etld_lib_functions.logger.error(f"Issue with xml file: {xml_file_path}")
        exit(1)


def host_list_detection_05_update_knowledgebase(refresh_knowledgebase_flag=True):
    begin_host_list_detection_05_update_knowledgebase()

    try:
        update_knowledgebase_in_host_list_detection_database(refresh_knowledgebase_flag=refresh_knowledgebase_flag)
    except Exception as e:
        etld_lib_functions.logger.error(f"Exception: {e}")
        etld_lib_functions.logger.error(f"Issue updating knowledgebase")
        exit(1)

    end_host_list_detection_05_update_knowledgebase()


def host_list_detection_05_final_validation():
    sqlite_file = etld_lib_config.host_list_detection_sqlite_file
    try:
        sqlite_obj = etld_lib_sqlite_tables.SqliteObj(sqlite_file=sqlite_file)
        sqlite_obj.validate_all_tables_loaded_successfully()
        sqlite_obj.close_connection()

    except Exception as e:
        etld_lib_functions.logger.error(f"Exception: {e}")
        etld_lib_functions.logger.error(f"Issue updating final validation for {sqlite_file}")
        exit(1)


def end_message_info():
    xml_file_list = sorted(
        Path(etld_lib_config.host_list_detection_extract_dir).glob(
            etld_lib_config.host_list_detection_extract_dir_file_search_blob))
    for host_list_detection_xml_file in xml_file_list:
        if str(host_list_detection_xml_file).endswith('.xml') or str(host_list_detection_xml_file).endswith('.xml.gz'):
            etld_lib_functions.log_file_info(host_list_detection_xml_file, 'input file')


def end_host_list_detection_05_transform_load():
    end_message_info()
    etld_lib_functions.logger.info(f"end")


def begin_host_list_detection_05_transform_load():
    etld_lib_functions.logger.info("start")


def end_host_list_detection_05_update_knowledgebase():
    etld_lib_functions.logger.info(f"end")


def begin_host_list_detection_05_update_knowledgebase():
    etld_lib_functions.logger.info("start")


def main(multiprocessing_flag=False, queue_of_file_paths: Queue = Queue(), run_transform=True,
         refresh_knowledgebase_flag=True):
    # Multiprocessing is executed through spawn_multiprocessing_queue_to_transform_and_load_xml_files_into_sqlite()
    # Toggle run_transform and refresh_knowledgebase_flag for testing.
    if run_transform:
        host_list_detection_transform_and_load_all_xml_files_into_sqlite(multiprocessing_flag=multiprocessing_flag,
                                                                         queue_of_file_paths=queue_of_file_paths)
    host_list_detection_05_update_knowledgebase(refresh_knowledgebase_flag)
    update_final_status_in_host_list_detection_database()
    host_list_detection_05_final_validation()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='host_list_detection_05_transform_load_xml_to_sqlite')
    etld_lib_config.main()
    # etld_lib_credentials.main()
    etld_lib_authentication_objects.main()
    main(multiprocessing_flag=False, queue_of_file_paths=Queue(), run_transform=True, refresh_knowledgebase_flag=False)
    # Toggle run_transform and refresh_knowledgebase_flag for testing.



#!/usr/bin/env python3
import xmltodict
import json
import re
from pathlib import Path
import time

from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_sqlite_tables
from qualys_etl.etld_lib import etld_lib_datetime
from qualys_etl.etld_lib import etld_lib_extract_transform_load
import codecs


def create_counter_objects(database_type='sqlite', table_name=""):
    counter_obj_kb = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=25000,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"rows added/updated into {database_type} "
                                    f"table {table_name}")

    counter_obj_kb_duplicates = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=25000,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"duplicate rows not added to {database_type} "
                                    f"table {table_name}")

    counter_obj_dict_new = {'counter_obj_kb': counter_obj_kb,
                            'counter_obj_kb_duplicates': counter_obj_kb_duplicates}

    return counter_obj_dict_new


def drop_and_create_temp_merge_tables(sqlite_obj):
    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.kb_table_name_merge_new_data,
        csv_columns=etld_lib_config.kb_csv_columns(),
        csv_column_types=etld_lib_config.kb_csv_column_types(),
        key=['QID'])

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.kb_status_table_name,
        csv_columns=etld_lib_config.status_table_csv_columns(),
        csv_column_types=etld_lib_config.status_table_csv_column_types(),
        key=['STATUS_NAME'])


def drop_and_create_status_table(sqlite_obj):
    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.kb_status_table_name,
        csv_columns=etld_lib_config.status_table_csv_columns(),
        csv_column_types=etld_lib_config.status_table_csv_column_types(),
        key=['STATUS_NAME'])


def drop_and_create_all_tables(sqlite_obj):
    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.kb_table_name,
        csv_columns=etld_lib_config.kb_csv_columns(),
        csv_column_types=etld_lib_config.kb_csv_column_types(),
        key=['QID'])

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.kb_status_table_name,
        csv_columns=etld_lib_config.status_table_csv_columns(),
        csv_column_types=etld_lib_config.status_table_csv_column_types(),
        key=['STATUS_NAME'])


def drop_and_create_all_views(sqlite_obj):
    drop_view = f"DROP VIEW IF EXISTS {etld_lib_config.kb_table_name_cve_list_view}"
    sqlite_obj.execute_statement(drop_view)
    create_view_select_statement = '''select SUBSTR(CVE_LIST_ITEM_PREFIX, 0, CVE_LIST_ITEM_PREFIX_LENGTH) AS CVE, * from (
select Q_KnowledgeBase.*, INSTR(json_each.value,'CVE') as BEGIN_CVE_LIST_ITEM, json_each.value as CVE_LIST_ITEM_JSON, 
SUBSTR(json_each.value, INSTR(json_each.value,'CVE')) AS CVE_LIST_ITEM_PREFIX,
INSTR(SUBSTR(json_each.value, INSTR(json_each.value,'CVE')),'","') as CVE_LIST_ITEM_PREFIX_LENGTH 
from Q_KnowledgeBase, json_each( Q_KnowledgeBase.CVE_LIST, '$.CVE' )
where CVE_LIST like '%{%'
) 
ORDER BY CVE DESC'''

    create_view = \
        f"CREATE VIEW {etld_lib_config.kb_table_name_cve_list_view} as " \
        f"{create_view_select_statement}"
    sqlite_obj.execute_statement(create_view)


def insert_one_row_into_table(
        item_dict: dict,
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
        table_name: str,
        counter_obj: dict,
        batch_date: str,
        batch_number: str

):

    # def prepare_field(field_data: dict, field_name_tmp):
    #     if field_data is None:
    #         field_data = ""
    #     elif 'DATE' in field_name_tmp:
    #         field_data = field_data.replace("T", " ").replace("Z", "")
    #         field_data = re.sub("\\..*$", "", field_data)
    #     elif isinstance(field_data, int):
    #         field_data = str(field_data)
    #     elif 'CVE_LIST' in field_name_tmp:
    #         if 'CVE' in field_data.keys():
    #             if isinstance(field_data['CVE'], dict):
    #                 one_item_dict = [field_data['CVE']]
    #                 field_data['CVE'] = one_item_dict
    #                 field_data = json.dumps(field_data)
    #             else:
    #                 field_data = json.dumps(field_data)
    #     elif not isinstance(field_data, str):
    #         field_data = json.dumps(field_data)
    #
    #     return field_data

    # row_in_sqlite_form = []
    #
    # for field_name in etld_lib_config.kb_csv_columns():  # Iterate through expected columns (contract)
    #     if field_name in item_dict.keys():  # Iterate through columns found in dictionary
    #         item_dict[field_name] = \
    #             sqlite_obj.prepare_knowledgebase_database_field(
    #                 item_dict[field_name],
    #                 field_name,
    #                 etld_lib_config.kb_csv_column_types())
    #         row_in_sqlite_form.append(item_dict[field_name])
    #     else:
    #         row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field

    row_in_sqlite_form = \
        sqlite_obj.prepare_database_row_vmpc(
            item_dict,
            etld_lib_config.kb_csv_columns(),
            etld_lib_config.kb_csv_column_types(), batch_date=batch_date, batch_number=batch_number)

    result = sqlite_obj.insert_or_replace_row_pristine(table_name, row_in_sqlite_form)

    if result is True:
        counter_obj['counter_obj_kb'].update_counter_and_display_to_log()
    else:
        counter_obj['counter_obj_kb_duplicates'].update_counter_and_display_to_log()


def insert_xml_file_into_sqlite(xml_file: Path,
                                sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
                                table_name: str,
                                counter_obj: dict,
                                compression_method=open):

    def callback_to_insert_host_into_sqlite(element_names: tuple, document_item: dict):
        if len(element_names) > 2 and "VULN" != element_names[3][0]:
            return True
        # document_item['BATCH_DATE'] = batch_date
        # document_item['BATCH_NUMBER'] = int(batch_number)
        #document_item['Row_Last_Updated'] = etld_lib_datetime.get_utc_datetime_sqlite_database_format()
        insert_one_row_into_table(
            document_item, sqlite_obj, table_name, counter_obj=counter_obj,
            batch_date=batch_date, batch_number=batch_number)
        return True

    if etld_lib_config.xmltodict_parse_using_codec_to_replace_utf8_error is False:
        with compression_method(str(xml_file), "rt", encoding='utf-8') as xml_file_fd:
            batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(xml_file)
            batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(xml_file)
            xmltodict.parse(xml_file_fd.read(),
                            item_depth=4,
                            item_callback=callback_to_insert_host_into_sqlite)
    else:
        with compression_method(str(xml_file), "rb") as xml_file_fd:
            batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(xml_file)
            batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(xml_file)
            xmltodict.parse(codecs.decode(xml_file_fd.read(), encoding='utf-8', errors='replace'),
                            item_depth=4,
                            item_callback=callback_to_insert_host_into_sqlite)


def kb_transform_and_load_all_xml_files_into_sqlite():
    # TODO Drop Merge Table if exists, load new data into merge table,
    #  if not 1970, load data from old table where not exists into merge table,
    #  rename merge table to old table name.
    begin_knowledgebase_05_tranform_load()
    xml_file_to_import_into_sqlite = ""
    try:
        kb_sqlite_obj = etld_lib_sqlite_tables.SqliteObj(sqlite_file=etld_lib_config.kb_sqlite_file)
        xml_file_list = []
        rebuild_flag = False
        for file_name in sorted(Path(etld_lib_config.kb_extract_dir).glob(etld_lib_config.kb_extract_dir_file_search_blob)):
            if str(file_name).endswith('.xml') or str(file_name).endswith('.xml.gz'):
                xml_file_list.append(file_name)

        xml_file_to_import_into_sqlite = ""
        if len(xml_file_list) > 0:
            xml_file_to_import_into_sqlite = xml_file_list[-1]
            if xml_file_to_import_into_sqlite.name.__contains__("1970"):
                rebuild_flag = True
        else:
            etld_lib_functions.logger.error("No xml files to process, rerun extract when ready.")
            etld_lib_functions.logger.error(f"Directory: {etld_lib_config.kb_extract_dir}")
            exit(1)

        drop_and_create_status_table(sqlite_obj=kb_sqlite_obj)

        if rebuild_flag is True:
            # REBUILD KNOWLEDGEBASE
            counter_obj_dict = create_counter_objects(table_name=etld_lib_config.kb_table_name)
            drop_and_create_all_tables(sqlite_obj=kb_sqlite_obj)
            drop_and_create_all_views(sqlite_obj=kb_sqlite_obj)
            insert_xml_file_into_sqlite(
                xml_file=xml_file_to_import_into_sqlite,
                sqlite_obj=kb_sqlite_obj,
                table_name=etld_lib_config.kb_table_name,
                counter_obj=counter_obj_dict,
                compression_method=etld_lib_config.kb_open_file_compression_method)
        else:
            counter_obj_dict = create_counter_objects(table_name=etld_lib_config.kb_table_name)
            drop_and_create_all_views(sqlite_obj=kb_sqlite_obj)
            insert_xml_file_into_sqlite(
                xml_file=xml_file_to_import_into_sqlite,
                sqlite_obj=kb_sqlite_obj,
                table_name=etld_lib_config.kb_table_name,
                counter_obj=counter_obj_dict,
                compression_method=etld_lib_config.kb_open_file_compression_method)
        batch_name = etld_lib_extract_transform_load.get_batch_name_from_filename(
            file_name=xml_file_to_import_into_sqlite)
        batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(
            file_name=xml_file_to_import_into_sqlite)
        batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(
            file_name=xml_file_to_import_into_sqlite)
        etld_lib_functions.logger.info(f"Received batch file from multiprocessing Queue: {batch_name}")

        kb_sqlite_obj.update_status_table(
                batch_date=batch_date, batch_number=batch_number,
                total_rows_added_to_database=counter_obj_dict['counter_obj_kb'].get_counter(),
                status_table_name=etld_lib_config.kb_status_table_name,
                status_table_columns=etld_lib_config.status_table_csv_columns(),
                status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                status_name_column=etld_lib_config.kb_table_name, status_column='final')

        kb_sqlite_obj.update_status_table(
            batch_date=batch_date,
            batch_number=0,
            total_rows_added_to_database=0,
            status_table_name=etld_lib_config.kb_status_table_name,
            status_table_columns=etld_lib_config.status_table_csv_columns(),
            status_table_column_types=etld_lib_config.status_table_csv_column_types(),
            status_name_column='ALL_TABLES_LOADED_SUCCESSFULLY',
            status_column='final')

        kb_sqlite_obj.commit_changes()
        etld_lib_functions.log_file_info(xml_file_to_import_into_sqlite, "File loaded into database")
        kb_sqlite_obj.vacuum_database()
        kb_sqlite_obj.commit_changes()
        etld_lib_functions.log_file_info(etld_lib_config.kb_sqlite_file, "Database Vacuum Completed")
        kb_sqlite_obj.validate_all_tables_loaded_successfully()
        kb_sqlite_obj.close_connection()
        counter_obj_dict['counter_obj_kb'].display_final_counter_to_log()
        end_knowledgebase_05_transform_load()

    except Exception as e:
        etld_lib_functions.logger.warning(f"Exception: {e}")
        etld_lib_functions.logger.warning(f"Issue with xml file: {xml_file_to_import_into_sqlite}")
        raise Exception("Raise exception warning processing knowledgebase for retry attempt.")


def end_knowledgebase_05_transform_load():
    etld_lib_functions.logger.info(f"end")


def begin_knowledgebase_05_tranform_load():
    etld_lib_functions.logger.info("start")


def main():
    max_retry_count = 5
    retry_count = 0
    sleep_seconds = 300
    while retry_count < max_retry_count:
        try:
            kb_transform_and_load_all_xml_files_into_sqlite()
            break  # Success
        except Exception as e:
            retry_count += 1
            etld_lib_functions.logger.warning(f"Retry kb_transform_and_load_all_xml_files_into_sqlite")
            etld_lib_functions.logger.warning(f"Retry number: {retry_count} of max_retry_count: {max_retry_count}")
            etld_lib_functions.logger.warning(f"Sleeping {sleep_seconds} before retry...")
            time.sleep(sleep_seconds)
    else:
        raise Exception("Retries for processing knowledgebase failed.")


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='knowledgebase_05_transform_load_xml_to_sqlite')
    etld_lib_config.main()
    main()



#!/usr/bin/env python3
from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_csv_distribution


def begin_asset_inventory_06_distribution():
    etld_lib_functions.logger.info(f"start")


def end_asset_inventory_06_distribution():
    etld_lib_functions.logger.info(f"end")


def extract_csv_from_database(test_extract=False):
    etld_lib_functions.logger.info(f"test_extract={test_extract}, "
                                   f"test_system={etld_lib_config.test_system_do_not_test_intermediary_extracts_flag}")
    if test_extract or etld_lib_config.test_system_do_not_test_intermediary_extracts_flag:
        etld_lib_csv_distribution.distribute_csv_data_for_one_workflow(
            distribution_csv_flag=etld_lib_config.asset_inventory_distribution_csv_flag,
            etl_workflow='asset_inventory_06_distribution',
            distribution_csv_flag_name='asset_inventory_distribution_csv_flag'
        )
    else:
        etld_lib_csv_distribution.distribute_csv_data_for_one_workflow(
            distribution_csv_flag=etld_lib_config.asset_inventory_distribution_csv_flag,
            etl_workflow='asset_inventory_etl_workflow',
            distribution_csv_flag_name='asset_inventory_distribution_csv_flag',
            exclude_table_names=['Q_Asset_Inventory']
        )


def main(test_extract=False):
    begin_asset_inventory_06_distribution()
    extract_csv_from_database(test_extract)
    end_asset_inventory_06_distribution()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='asset_inventory_06_distribution')
    etld_lib_config.main()
    main(test_extract=True)



#!/usr/bin/env python3
import json
from pathlib import Path
import re
from multiprocessing import Process, Queue
import time

from qualys_etl.etld_lib import etld_lib_config
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_extract_transform_load
from qualys_etl.etld_lib import etld_lib_sqlite_tables


def insert_one_row_into_was_table(data_list_item, sqlite_obj, counter_objects, batch_name, batch_date, batch_number):

    if 'WebApp' in data_list_item:
        insert_one_row_into_table(
            sqlite_obj=sqlite_obj,
            table_name=etld_lib_config.was_webapp_table_name,
            table_fields=etld_lib_config.was_webapp_csv_columns(),
            table_field_types=etld_lib_config.was_webapp_csv_column_types(),
            batch_name=batch_name,
            batch_date=batch_date,
            batch_number=batch_number,
            item_dict=data_list_item['WebApp'],
            counter_obj=counter_objects['counter_obj_was_webapp'],
            counter_obj_duplicates=counter_objects['counter_obj_was_webapp_duplicates']
        )
    elif 'Finding' in data_list_item:
        insert_one_row_into_table(
            sqlite_obj=sqlite_obj,
            table_name=etld_lib_config.was_finding_table_name,
            table_fields=etld_lib_config.was_finding_csv_columns(),
            table_field_types=etld_lib_config.was_finding_csv_column_types(),
            batch_name=batch_name,
            batch_date=batch_date,
            batch_number=batch_number,
            item_dict=data_list_item['Finding'],
            counter_obj=counter_objects['counter_obj_was_finding'],
            counter_obj_duplicates=counter_objects['counter_obj_was_finding_duplicates']
        )
    elif 'Catalog' in data_list_item:
        insert_one_row_into_table(
            sqlite_obj=sqlite_obj,
            table_name=etld_lib_config.was_catalog_table_name,
            table_fields=etld_lib_config.was_catalog_csv_columns(),
            table_field_types=etld_lib_config.was_catalog_csv_column_types(),
            batch_name=batch_name,
            batch_date=batch_date,
            batch_number=batch_number,
            item_dict=data_list_item['Catalog'],
            counter_obj=counter_objects['counter_obj_was_catalog'],
            counter_obj_duplicates=counter_objects['counter_obj_was_catalog_duplicates']
        )


def prepare_was_finding_webapp_fields(item_dict):
    item_dict['webApp_id'] = 0
    item_dict['webApp_name'] = 'NONE'
    item_dict['webApp_url'] = 'NONE'
    item_dict['webApp_tags'] = '{"list": [], "count": 0}}'

    if 'webApp' in item_dict:
        if 'id' in item_dict['webApp']:
            item_dict['webApp_id'] = item_dict['webApp']['id']
        if 'tags' in item_dict['webApp']:
            item_dict['webApp_tags'] = item_dict['webApp']['tags']
        if 'url' in item_dict['webApp']:
            item_dict['webApp_url'] = item_dict['webApp']['url']
        if 'name' in item_dict['webApp']:
            item_dict['webApp_name'] = item_dict['webApp']['name']
        if 'group' in item_dict:
            item_dict['group_0'] = item_dict['group']
        if 'function' in item_dict:
            item_dict['function_0'] = item_dict['function']

    return item_dict


def insert_one_row_into_table(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj,
        table_name: str,
        table_fields: list,
        table_field_types: list,
        batch_name: str,
        batch_date: str,
        batch_number: str,
        item_dict: dict,
        counter_obj,
        counter_obj_duplicates):

    # def prepare_was_field(field_data, field_name_tmp):
    #     if field_data is None:
    #         field_data = ""
    #     elif 'Date' in field_name_tmp:
    #         field_data = field_data.replace("T", " ").replace("Z", "")
    #         field_data = re.sub("\\..*$", "", field_data)
    #     elif 'lastBoot' in field_name_tmp:
    #         field_data = field_data.replace("T", " ").replace("Z", "")
    #         field_data = re.sub("\\..*$", "", field_data)
    #     elif isinstance(field_data, int):
    #         field_data = str(field_data)
    #     elif not isinstance(field_data, str):
    #         field_data = json.dumps(field_data)
    #
    #     return field_data

    # if Finding.webApp (webApp)
    # flatten webApp data and reinsert into item_dict?
    if 'webApp' in item_dict:
        item_dict = prepare_was_finding_webapp_fields(item_dict)

    row_in_sqlite_form = sqlite_obj.prepare_database_row_was(
        item_dict=item_dict,
        csv_columns=table_fields,
        csv_column_types=table_field_types,
        batch_name=batch_name,
        batch_date=batch_date,
        batch_number=batch_number
    )

    # for field_name in table_fields:  # Iterate through expected columns (contract)
    #     if field_name in item_dict.keys():  # Iterate through columns found in dictionary
    #         item_dict[field_name] = \
    #             prepare_was_field(item_dict[field_name], field_name)
    #         row_in_sqlite_form.append(item_dict[field_name])
    #     else:
    #         row_in_sqlite_form.append("")  # Ensure blank is added to each required empty field

    result = sqlite_obj.insert_unique_row_ignore_duplicates(table_name, row_in_sqlite_form)
    if result is True:
        counter_obj.update_counter_and_display_to_log()
    else:
        counter_obj_duplicates.update_counter_and_display_to_log()


def drop_and_create_all_tables(sqlite_obj):
    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.was_status_table_name,
        csv_columns=etld_lib_config.status_table_csv_columns(),
        csv_column_types=etld_lib_config.status_table_csv_column_types(),
        key='STATUS_NAME')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.was_webapp_table_name,
        csv_columns=etld_lib_config.was_webapp_csv_columns(),
        csv_column_types=etld_lib_config.was_webapp_csv_column_types(),
        key='id')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.was_catalog_table_name,
        csv_columns=etld_lib_config.was_catalog_csv_columns(),
        csv_column_types=etld_lib_config.was_catalog_csv_column_types(),
        key='id')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.was_finding_table_name,
        csv_columns=etld_lib_config.was_finding_csv_columns(),
        csv_column_types=etld_lib_config.was_finding_csv_column_types(),
        key='id')

    sqlite_obj.drop_and_recreate_table(
        table_name=etld_lib_config.was_q_knowledgebase_in_was_finding_table,
        csv_columns=etld_lib_config.kb_csv_columns(),
        csv_column_types=etld_lib_config.kb_csv_column_types(),
        key='QID')


def create_counter_objects(database_type='sqlite'):

    counter_obj_was_webapp = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=100,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"rows added to {database_type} "
                                    f"table {etld_lib_config.was_webapp_table_name}")

    counter_obj_was_catalog = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=100,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"rows added to {database_type} "
                                    f"table {etld_lib_config.was_catalog_table_name}")

    counter_obj_was_finding = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=10000,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"rows added to {database_type} "
                                    f"table {etld_lib_config.was_finding_table_name}")

    counter_obj_was_webapp_duplicates = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=100,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"duplicate rows bypassed, not written to {database_type} "
                                    f"table {etld_lib_config.was_webapp_table_name}")

    counter_obj_was_catalog_duplicates = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=100,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"duplicate rows bypassed, not written to {database_type} "
                                    f"table {etld_lib_config.was_catalog_table_name}")

    counter_obj_was_finding_duplicates = etld_lib_functions.DisplayCounterToLog(
        display_counter_at=10000,
        logger_func=etld_lib_functions.logger.info,
        display_counter_log_message=f"duplicate rows bypassed, not written to {database_type} "
                                    f"table {etld_lib_config.was_finding_table_name}")

    counter_objects = {'counter_obj_was_webapp': counter_obj_was_webapp,
                       'counter_obj_was_finding': counter_obj_was_finding,
                       'counter_obj_was_catalog': counter_obj_was_catalog,
                       'counter_obj_was_webapp_duplicates': counter_obj_was_webapp_duplicates,
                       'counter_obj_was_catalog_duplicates': counter_obj_was_catalog_duplicates,
                       'counter_obj_was_finding_duplicates': counter_obj_was_finding_duplicates}

    return counter_objects


# def transform_and_load_queue_of_json_files_into_sqlite(queue_of_json_files,
#                                                        table_name,
#                                                        table_columns,
#                                                        table_column_types,
#                                                        table_primary_keys):
#     file_path = "EXCEPTION"
#     try:
#         sqlite_was = etld_lib_sqlite_tables.SqliteObj(etld_lib_config.was_sqlite_file)
#         drop_and_create_all_tables(sqlite_was)
#         counter_objects = create_counter_objects()
#         while True:
#             time.sleep(2)
#             file_path = queue_of_json_files.get()
#             if file_path == 'BEGIN':
#                 etld_lib_functions.logger.info(f"Found BEGIN of Queue.")
#                 break
#
#         while True:
#             time.sleep(2)
#             file_path = queue_of_json_files.get()
#             if file_path == 'END':
#                 etld_lib_functions.logger.info(f"Found END of Queue.")
#                 break
#
#             with etld_lib_config.was_open_file_compression_method(str(file_path), "rt", encoding='utf-8') \
#                     as read_file:
#                 batch_name = etld_lib_extract_transform_load.get_batch_name_from_filename(file_name=file_path)
#                 batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(file_name=file_path)
#                 batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(file_name=file_path)
#                 etld_lib_functions.logger.info(f"Received batch file from multiprocessing Queue: {batch_name}")
#
#                 status_name, status_detail_dict, status_count = \
#                     sqlite_was.update_status_table(
#                         batch_date=batch_date, batch_number=batch_number,
#                         total_rows_added_to_database=counter_objects['counter_obj_was'].get_counter(),
#                         etl_workflow_status_table_name=etld_lib_config.was_status_table_name,
#                         status_table_columns=etld_lib_config.status_table_csv_columns(),
#                         status_table_column_types=etld_lib_config.status_table_csv_column_types(),
#                         status_name_column='ASSET_INVENTORY_LOAD_STATUS', status_column='begin')
#
#                 all_items = json.load(read_file)
#                 for item in all_items['assetListData']['asset']:
#                     insert_one_asset_into_multiple_tables(item, sqlite_was, counter_objects)
#                 sqlite_was.commit_changes()
#
#
#                 status_name, status_detail_dict, status_count = \
#                     sqlite_was.update_status_table(
#                         batch_date=batch_date, batch_number=batch_number,
#                         total_rows_added_to_database=counter_objects['counter_obj_was'].get_counter(),
#                         etl_workflow_status_table_name=etld_lib_config.was_status_table_name,
#                         status_table_columns=etld_lib_config.status_table_csv_columns(),
#                         status_table_column_types=etld_lib_config.status_table_csv_column_types(),
#                         status_name_column='ASSET_INVENTORY_LOAD_STATUS', status_column='end')
#
#                 etld_lib_functions.logger.info(
#                     f"Committed batch file from multiprocessing Queue into Database: {batch_name}")
#
#         etld_lib_functions.logger.info(f"Completed processing Queue of files")
#
#         status_name, status_detail_dict, status_count = \
#             sqlite_was.update_status_table(
#                 batch_date=batch_date, batch_number=batch_number,
#                 total_rows_added_to_database=counter_objects['counter_obj_was'].get_counter(),
#                 etl_workflow_status_table_name=etld_lib_config.was_status_table_name,
#                 status_table_columns=etld_lib_config.status_table_csv_columns(),
#                 status_table_column_types=etld_lib_config.status_table_csv_column_types(),
#                 status_name_column='ASSET_INVENTORY_LOAD_STATUS', status_column='final')
#
#         sqlite_was.commit_changes()
#         sqlite_was.close_connection()
#         counter_objects['counter_obj_was_duplicates'].display_final_counter_to_log()
#         counter_objects['counter_obj_was'].display_final_counter_to_log()
#         counter_objects['counter_obj_software_os'].display_final_counter_to_log()
#         counter_objects['counter_obj_software_assetid'].display_final_counter_to_log()
#         end_msg_was_to_sqlite()
#
#     except Exception as e:
#         etld_lib_functions.logger.error(f"Exception: {e}")
#         etld_lib_functions.logger.error(f"Potential JSON File corruption detected: {file_path}")
#         exit(1)
#

def end_was_05_transform_load():
    etld_lib_functions.logger.info(f"end")


def begin_was_05_transform_load():
    etld_lib_functions.logger.info("start")


def attach_database(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj, table_name, database_as_name, database_file):
    sqlite_obj.attach_database_to_connection(database_as_name=database_as_name,
                                             database_sqlite_file=database_file)
    table_name = f"{database_as_name}.{table_name}"
    table_columns = sqlite_obj.get_table_columns(table_name=table_name)
    etld_lib_functions.logger.info(f"Found Table {table_name} columns: {table_columns}")
    return table_name, table_columns


def copy_database_table(
        sqlite_obj: etld_lib_sqlite_tables.SqliteObj, table_name, new_table_name, where_clause=""):
    try:
        etld_lib_functions.logger.info(f"Begin creating {new_table_name} from {table_name}")
        #        etl_workflow_sqlite_obj.cursor.execute(f"DROP TABLE IF EXISTS {new_table_name}")
        #        etl_workflow_sqlite_obj.cursor.execute(f"CREATE TABLE {new_table_name} AS SELECT * FROM {table_name} where 1=0")
        sqlite_obj.cursor.execute(f"INSERT INTO {new_table_name} SELECT * FROM {table_name} {where_clause}")
        etld_lib_functions.logger.info(f"End   creating {new_table_name} from {table_name}")
    except Exception as e:
        etld_lib_functions.logger.error(f"Error creating table {new_table_name} from {table_name}")
        etld_lib_functions.logger.error(f"Exception is: {e}")
        exit(1)


def create_kb_table_in_was_database(sqlite_obj: etld_lib_sqlite_tables.SqliteObj):
    etld_lib_functions.logger.info(f"Attaching database: {etld_lib_config.kb_table_name}")
    kb_table_name, kb_table_columns = \
        attach_database(sqlite_obj=sqlite_obj,
                        table_name="Q_KnowledgeBase",
                        database_as_name="K1",
                        database_file=etld_lib_config.kb_sqlite_file)

    etld_lib_functions.logger.info(f"Copying table: {kb_table_name}")
    where_clause = f"where {kb_table_name}.QID in (select distinct(qid) from Q_Was_Finding)"

    copy_database_table(sqlite_obj=sqlite_obj, table_name=kb_table_name,
                        new_table_name=etld_lib_config.was_q_knowledgebase_in_was_finding_table,
                        where_clause=where_clause)
    etld_lib_functions.logger.info(f"Commit copy of table: {kb_table_name}")
    sqlite_obj.commit_changes()


def transform_and_load_all_json_files_into_sqlite(dir_file_search_blob=None):
    json_file = "EXCEPTION"
    begin_was_05_transform_load()
    try:
        sqlite_was = etld_lib_sqlite_tables.SqliteObj(etld_lib_config.was_sqlite_file)
        drop_and_create_all_tables(sqlite_was)
        counter_objects = create_counter_objects()
        json_file_list = sorted(Path(etld_lib_config.was_extract_dir).
                                glob(etld_lib_config.was_extract_dir_file_search_blob_webapp_detail))
        json_file_list = json_file_list + sorted(Path(etld_lib_config.was_extract_dir).
                                                 glob(etld_lib_config.was_extract_dir_file_search_blob_finding_detail))
        json_file_list = json_file_list + sorted(Path(etld_lib_config.was_extract_dir).
                                                 glob(etld_lib_config.was_extract_dir_file_search_blob_catalog))

        for json_file in json_file_list:
            with etld_lib_config.was_open_file_compression_method(str(json_file), "rt", encoding='utf-8') \
                    as read_file:

                batch_name = etld_lib_extract_transform_load.get_batch_name_from_filename(file_name=json_file)
                batch_number = etld_lib_extract_transform_load.get_batch_number_from_filename(file_name=json_file)
                batch_date = etld_lib_extract_transform_load.get_batch_date_from_filename(file_name=json_file)
                etld_lib_functions.logger.info(f"Received batch file from directory: {batch_name}")
                status_table_columns = etld_lib_config.status_table_csv_columns()
                status_table_column_types = etld_lib_config.status_table_csv_column_types()

                if json_file.name.__contains__("was_webapp_detail_utc"):
                    total_rows_added_to_database = counter_objects['counter_obj_was_webapp'].get_counter()
                    status_name_column = etld_lib_config.was_webapp_table_name
                    counter_objects['counter_obj_was_webapp'].update_batch_info(
                        batch_name=batch_name,
                        batch_number=batch_number,
                        batch_date=batch_date,
                        status_table_name=etld_lib_config.was_status_table_name,
                        status_name_column=status_name_column)
                elif json_file.name.__contains__("was_finding_detail_utc"):
                    total_rows_added_to_database = counter_objects['counter_obj_was_finding'].get_counter()
                    status_name_column = etld_lib_config.was_finding_table_name
                    counter_objects['counter_obj_was_finding'].update_batch_info(
                        batch_name=batch_name,
                        batch_number=batch_number,
                        batch_date=batch_date,
                        status_table_name=etld_lib_config.was_status_table_name,
                        status_name_column=status_name_column)
                elif json_file.name.__contains__("was_catalog_utc"):
                    total_rows_added_to_database = counter_objects['counter_obj_was_catalog'].get_counter()
                    status_name_column = etld_lib_config.was_catalog_table_name
                    counter_objects['counter_obj_was_catalog'].update_batch_info(
                        batch_name=batch_name,
                        batch_number=batch_number,
                        batch_date=batch_date,
                        status_table_name=etld_lib_config.was_status_table_name,
                        status_name_column=status_name_column)
                else:
                    # skip all other files
                    continue

                sqlite_was.update_status_table(
                        batch_date=batch_date,
                        batch_number=batch_number,
                        total_rows_added_to_database=total_rows_added_to_database,
                        status_table_name=etld_lib_config.was_status_table_name,
                        status_table_columns=etld_lib_config.status_table_csv_columns(),
                        status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                        status_name_column=status_name_column,
                        status_column='begin')

                json_data = json.load(read_file)

                json_success_found = False
                if 'ServiceResponse' in json_data:
                    if 'responseCode' in json_data['ServiceResponse']:
                        responseCode = json_data['ServiceResponse']['responseCode']
                        if 'success' == str(responseCode).lower():
                            json_success_found = True

                if json_success_found:
                    pass
                else:
                    etld_lib_functions.logger.error(
                        f"C1 - Found invalid batch json with no responseCode SUCCESS: {status_name_column} - {batch_name}")
                    json_data_string = str(json.dumps(json_data)).replace('\n', ' ').replace('\r', ' ')
                    etld_lib_functions.logger.error(f"JSON DATA: {json_data_string}")
                    etld_lib_functions.logger.error(f"json_file: {json_file}")
                    exit(1)

                if 'data' in json_data['ServiceResponse']:
                    for data_list_item in json_data['ServiceResponse']['data']:
                        insert_one_row_into_was_table(data_list_item, sqlite_was, counter_objects, batch_name, batch_date, batch_number)
                    sqlite_was.commit_changes()
                    etld_lib_functions.logger.info(
                        f"Added batch file from directory into Database: {status_name_column} - {batch_name}")

                    sqlite_was.update_status_table(
                            batch_date=batch_date,
                            batch_number=batch_number,
                            total_rows_added_to_database=total_rows_added_to_database,
                            status_table_name=etld_lib_config.was_status_table_name,
                            status_table_columns=etld_lib_config.status_table_csv_columns(),
                            status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                            status_name_column=status_name_column,
                            status_column='end')
                elif 'success' in str(json_data['ServiceResponse']['responseCode']).lower():
                    etld_lib_functions.logger.info(
                        f"Skipping, No Data in batch file from directory into "
                        f"Database: {status_name_column} - {batch_name}")
                else:
                    etld_lib_functions.logger.error(
                        f"C2 - Found invalid batch json with no responseCode SUCCESS: {status_name_column} - {batch_name}")
                    json_data_string = str(json.dumps(json_data)).replace('\n', ' ').replace('\r', ' ')
                    etld_lib_functions.logger.error(f"JSON DATA: {json_data_string}")
                    etld_lib_functions.logger.error(f"json_file: {json_file}")
                    exit(1)

        create_kb_table_in_was_database(sqlite_obj=sqlite_was)
        sqlite_was.update_status_table(
                batch_date=counter_objects['counter_obj_was_webapp'].get_batch_date(),
                batch_number=counter_objects['counter_obj_was_webapp'].get_batch_number(),
                total_rows_added_to_database=counter_objects['counter_obj_was_webapp'].get_counter(),
                status_table_name=etld_lib_config.was_status_table_name,
                status_table_columns=etld_lib_config.status_table_csv_columns(),
                status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                status_name_column=counter_objects['counter_obj_was_webapp'].get_status_name_column(),
                status_column='final')
        sqlite_was.update_status_table(
                batch_date=counter_objects['counter_obj_was_finding'].get_batch_date(),
                batch_number=counter_objects['counter_obj_was_finding'].get_batch_number(),
                total_rows_added_to_database=counter_objects['counter_obj_was_finding'].get_counter(),
                status_table_name=etld_lib_config.was_status_table_name,
                status_table_columns=etld_lib_config.status_table_csv_columns(),
                status_table_column_types=etld_lib_config.status_table_csv_column_types(),
                status_name_column=counter_objects['counter_obj_was_finding'].get_status_name_column(),
                status_column='final')
        sqlite_was.update_status_table(
            batch_date=counter_objects['counter_obj_was_catalog'].get_batch_date(),
            batch_number=counter_objects['counter_obj_was_catalog'].get_batch_number(),
            total_rows_added_to_database=counter_objects['counter_obj_was_catalog'].get_counter(),
            status_table_name=etld_lib_config.was_status_table_name,
            status_table_columns=etld_lib_config.status_table_csv_columns(),
            status_table_column_types=etld_lib_config.status_table_csv_column_types(),
            status_name_column=counter_objects['counter_obj_was_catalog'].get_status_name_column(),
            status_column='final')

        sqlite_was.update_status_table(
            batch_date=counter_objects['counter_obj_was_webapp'].get_batch_date(),
            batch_number=0,
            total_rows_added_to_database=0,
            status_table_name=etld_lib_config.was_status_table_name,
            status_table_columns=etld_lib_config.status_table_csv_columns(),
            status_table_column_types=etld_lib_config.status_table_csv_column_types(),
            status_name_column='ALL_TABLES_LOADED_SUCCESSFULLY',
            status_column='final')

        etld_lib_functions.logger.info(f"Completed processing directory of files")
        sqlite_was.commit_changes()
        sqlite_was.validate_all_tables_loaded_successfully()
        sqlite_was.close_connection()
        counter_objects['counter_obj_was_webapp'].display_final_counter_to_log()
        counter_objects['counter_obj_was_catalog'].display_final_counter_to_log()
        counter_objects['counter_obj_was_finding'].display_final_counter_to_log()

    except Exception as e:
        etld_lib_functions.logger.error(f"JSON File Issue: {json_file}")
        etld_lib_functions.logger.error(f"Exception: {e}")
        exit(1)
    end_was_05_transform_load()


# def spawn_multiprocessing_queue_to_transform_and_load_json_files_into_sqlite(file_type='was_webapp'):
#     queue_of_file_paths = Queue()
#     queue_process = \
#         Process(target=transform_and_load_queue_of_json_files_into_sqlite, args=(queue_of_file_paths,))
#     queue_process.daemon = True
#     queue_process.start()
#     batch_queue_of_file_paths = queue_of_file_paths
#     batch_queue_process = queue_process
#
#     return batch_queue_of_file_paths, batch_queue_process


def main():
    transform_and_load_all_json_files_into_sqlite()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='was_05_transform_load_json_to_sqlite')
    etld_lib_config.main()
    etld_lib_config.was_json_to_sqlite_via_multiprocessing = False
    main()



#!/usr/bin/env python3
import multiprocessing
import traceback
import time
import json
from pathlib import Path
from qualys_etl.etld_lib import etld_lib_config
#from qualys_etl.etld_lib import etld_lib_credentials
from qualys_etl.etld_lib import etld_lib_authentication_objects
from qualys_etl.etld_lib import etld_lib_functions
from qualys_etl.etld_lib import etld_lib_datetime
from qualys_etl.etld_lib import etld_lib_extract_transform_load
from qualys_etl.etld_pcrs import pcrs_05_transform_load_json_to_sqlite
from qualys_etl.etld_pcrs import pcrs_04_extract_from_qualys

global utc_datetime

def begin_pcrs_03_extract_controller(message=""):
    etld_lib_functions.logger.info(f"start {message}")


def end_pcrs_03_extract_controller(message=""):
    etld_lib_functions.logger.info(f"end  {message}")

def start_multiprocessing_transform_json_to_sqlite():

    if etld_lib_config.pcrs_json_to_sqlite_via_multiprocessing is True:
        batch_queue_of_file_paths, batch_queue_process = \
            pcrs_05_transform_load_json_to_sqlite.\
            spawn_multiprocessing_queue_to_transform_and_load_json_files_into_sqlite()
        etld_lib_functions.logger.info(f"Queue of json files process id: {batch_queue_process.pid} ")
        batch_queue_of_file_paths.put("BEGIN")
        return batch_queue_of_file_paths, batch_queue_process
    else:
        return None, None


def stop_multiprocessing_transform_json_to_sqlite(batch_queue_process, batch_queue_of_file_paths):

    if etld_lib_config.pcrs_json_to_sqlite_via_multiprocessing is True:
        batch_queue_of_file_paths.put("END")
        while True:
            if batch_queue_process.is_alive():
                etld_lib_functions.logger.info("Waiting for Queue to end.")
                time.sleep(15)
            else:
                etld_lib_functions.logger.info("Queue Completed.")
                break


def add_batch_to_transform_json_to_sqlite(batch_number_str,
                                          json_file,
                                          batch_queue_of_file_paths: multiprocessing.Queue,
                                          batch_queue_process: multiprocessing.Process):

    if etld_lib_config.pcrs_json_to_sqlite_via_multiprocessing is True:
        etld_lib_functions.logger.info(f"Sending batch file to multiprocessing Queue: {Path(json_file).name}")
        try:
            batch_queue_of_file_paths.put(str(json_file), timeout=30)
        except Exception as e:
            etld_lib_functions.logger.error(f"Problem putting file in queue: {str(json_file)}")
            exit(1)

        # if batch_queue_process.is_alive():
        #     pass
        # else:
        #     etld_lib_functions.logger.error("Batch Process was killed or database error, please investigate and retry.")
        #     exit(1)


def check_qualys_connection_rate_limits(batch_number_str, qualys_headers_dict, batch_info):

    if batch_number_str in qualys_headers_dict.keys():
        if 'x_ratelimit_remaining' in qualys_headers_dict[batch_number_str].keys():
            x_ratelimit_remaining = qualys_headers_dict[batch_number_str]['x_ratelimit_remaining']
            if int(x_ratelimit_remaining) < 15:
                etld_lib_functions.logger.warning(f"x_ratelimit_remaining is less than 15. "
                                                  f"Sleeping 5 min.  batch_info: {batch_info}, "
                                                  f"header_info: {qualys_headers_dict[batch_number_str]}")
                time.sleep(300)
        else:
            etld_lib_functions.logger.warning(f"x_ratelimit_remaining missing from Qualys Header. "
                                              f"Sleeping 5 min.  batch_info: {batch_info}, "
                                              f"header_info: {qualys_headers_dict[batch_number_str]}")
            time.sleep(300)


def remove_old_files():

    etld_lib_config.remove_old_files(
        dir_path=etld_lib_config.pcrs_extract_dir,
        dir_search_glob=etld_lib_config.pcrs_extract_dir_file_search_blob,
        other_files_list=etld_lib_config.pcrs_data_files
    )

    etld_lib_config.remove_old_files(
        dir_path=etld_lib_config.pcrs_distribution_dir,
        dir_search_glob=etld_lib_config.pcrs_distribution_dir_file_search_blob
    )


def get_pcrs_policy_list_data(batch_queue_process, batch_queue_of_file_paths):

    etld_lib_functions.logger.info("start get_pcrs_policy_list_data")
    qualys_headers_dict = {}
    batch_number = 0
    batch_number_str = f'batch_{batch_number:06d}'
    etld_lib_authentication_objects.qualys_authentication_obj.get_current_bearer_token()
    cred_dict = etld_lib_authentication_objects.qualys_authentication_obj.get_credentials_dict()

    file_info_dict = \
        etld_lib_config.prepare_extract_batch_file_name(
            next_batch_number_str='batch_000000',
            next_batch_date=utc_datetime,
            extract_dir=etld_lib_config.pcrs_extract_dir,
            file_name_type="pcrs_policy_list",
            file_name_option="lastEvaluationDate",
            file_name_option_date=etld_lib_config.pcrs_last_evaluation_date,
            file_extension="json",
            compression_method=etld_lib_config.pcrs_open_file_compression_method)

    policy_list = pcrs_04_extract_from_qualys.pcrs_extract_policy_list(
        last_evaluation_date=etld_lib_config.pcrs_last_evaluation_date,
        batch_number_str=batch_number_str,
        payload_dict=etld_lib_config.pcrs_policy_list_payload_option,
        qualys_headers_dict=qualys_headers_dict,
        cred_dict=cred_dict,
        file_info_dict=file_info_dict)

    if len(policy_list) > 0:
        add_batch_to_transform_json_to_sqlite(
            batch_number_str=batch_number_str, json_file=file_info_dict['next_file_path'],
            batch_queue_process=batch_queue_process, batch_queue_of_file_paths=batch_queue_of_file_paths)
    else:
        etld_lib_functions.logger.warning(f"Zero Policies Found Active")
        etld_lib_functions.logger.warning(f"Check permissions of Qualys User and determine if Policies are Active.")
        exit(0)

    etld_lib_functions.logger.info("end get_pcrs_policy_list_data")

def get_pcrs_hostids_data(batch_queue_process, batch_queue_of_file_paths):

    etld_lib_functions.logger.info("start get_pcrs_hostids_data")
    # Get Policy List
    file_list = etld_lib_config.get_list_of_matching_files(
        dir_path=etld_lib_config.pcrs_extract_dir,
        dir_search_glob=etld_lib_config.pcrs_extract_dir_file_search_blob_policy_list
    )
    if len(file_list) > 1:
        etld_lib_functions.logger.error(f"ERROR: Old pcrs_policy_list files exist in {etld_lib_config.pcrs_extract_dir}")
        etld_lib_functions.logger.error(f"ERROR: {file_list}")
        etld_lib_functions.logger.error(f"ERROR: please determine if files are locked due to permissions or other issues and rerun.")
        exit(1)
    elif len(file_list) != 1:
        etld_lib_functions.logger.error(f"ERROR: could not find pcrs_policy_list file in {etld_lib_config.pcrs_extract_dir}")
        etld_lib_functions.logger.error(f"ERROR: {file_list}")
        etld_lib_functions.logger.error(f"ERROR: please determine if your Qualys API User has access to PC Module.")
        exit(1)

    with etld_lib_config.pcrs_open_file_compression_method(file_list[0], "rt", encoding='utf-8') as read_file:
       policy_list = json.load(read_file)
       batch_number = 0
       active_policy_list = pcrs_04_extract_from_qualys.pcrs_get_sorted_active_policy_list(policy_list)
       for policy_item in active_policy_list:
           # Extract hostids for a policy.
           policy_id = policy_item['id']
           if get_pcrs_postureinfo_process_policy_id_status(policy_id):
               pass
           else:
               etld_lib_functions.logger.info(f"Bypassing Policy Id:{policy_id} due to include/exclude in etld_config_settings.yaml")
               continue

           batch_number = batch_number + 1
           batch_number_str = f'batch_{batch_number:06d}'
           batch_number_str = f'{batch_number_str}_policyId_{policy_id}'

           file_info_dict = \
               etld_lib_config.prepare_extract_batch_file_name(
                   next_batch_number_str=batch_number_str,
                   next_batch_date=utc_datetime,
                   extract_dir=etld_lib_config.pcrs_extract_dir,
                   file_name_type="pcrs_hostids",
                   file_name_option="lastEvaluationDate",
                   file_name_option_date=etld_lib_config.pcrs_last_evaluation_date,
                   file_extension="json",
                   compression_method=etld_lib_config.pcrs_open_file_compression_method)

           #cred_dict = etld_lib_credentials.get_bearer_stored_in_env(update_bearer=False, cred=cred_dict)
           etld_lib_authentication_objects.qualys_authentication_obj.get_current_bearer_token()
           cred_dict = etld_lib_authentication_objects.qualys_authentication_obj.get_credentials_dict()

           qualys_headers_dict = {}
           pcrs_04_extract_from_qualys.pcrs_extract_hostids_list(
               last_scan_date=etld_lib_config.pcrs_last_scan_date,
               last_evaluation_date=etld_lib_config.pcrs_last_evaluation_date,
               batch_number_str=batch_number_str,
               qualys_headers_dict=qualys_headers_dict,
               payload_dict=etld_lib_config.pcrs_hostids_payload_option,
               payload_postureinfo_dict=etld_lib_config.pcrs_postureinfo_payload_option,
               cred_dict=cred_dict,
               policy_id=policy_item['id'],
               file_info_dict=file_info_dict)

           hostids_list = pcrs_04_extract_from_qualys.pcrs_get_hostids_list(json_file=file_info_dict['next_file_path'])
           if len(hostids_list[0]['hostIds']) > 0:
               add_batch_to_transform_json_to_sqlite(
                   batch_number_str=batch_number_str, json_file=file_info_dict['next_file_path'],
                   batch_queue_process=batch_queue_process, batch_queue_of_file_paths=batch_queue_of_file_paths)
           else:
              etld_lib_functions.logger.info(f"PolicyID: {hostids_list[0]['policyId']}, Zero Hosts Found, not loaded to database")

       etld_lib_functions.logger.info("end   get_pcrs_hostids_data")


def get_pcrs_postureinfo_process_policy_id_status(pcrs_policy_id):
    process_policy_id_status = False

    pcrs_policy_id_include_list = etld_lib_config.pcrs_policy_id_include_list
    pcrs_policy_id_exclude_list = etld_lib_config.pcrs_policy_id_exclude_list
    if isinstance(pcrs_policy_id_include_list, list):
        if len(pcrs_policy_id_include_list) > 0:
            for policy_id in pcrs_policy_id_include_list:
                if str(policy_id) == str(pcrs_policy_id):
                    process_policy_id_status = True
                    break
        else:
            process_policy_id_status = True
    else:
        process_policy_id_status = True

    if isinstance(pcrs_policy_id_exclude_list, list):
        if len(pcrs_policy_id_exclude_list) > 0:
            for policy_id in pcrs_policy_id_exclude_list:
                if str(policy_id) == str(pcrs_policy_id):
                    process_policy_id_status = False
                    break

    return process_policy_id_status


def get_pcrs_postureinfo_data(batch_queue_process, batch_queue_of_file_paths):

    etld_lib_functions.logger.info("start get_pcrs_postureinfo_data")

    file_list = etld_lib_config.get_list_of_matching_files(
        dir_path=etld_lib_config.pcrs_extract_dir,
        dir_search_glob=etld_lib_config.pcrs_extract_dir_file_search_blob_hostids
    )
    # Loop through policy hostids files and extract postureinfo.
    for pcrs_hostids_file in file_list:
        get_pcrs_postureinfo_for_one_policy_with_multiprocessing(pcrs_hostids_file,
                                            batch_queue_process,
                                            batch_queue_of_file_paths,
                                            max_batch_size=etld_lib_config.pcrs_multi_proc_batch_size)

    etld_lib_functions.logger.info("end   get_pcrs_postureinfo_data")


def get_pcrs_postureinfo_for_one_policy_with_multiprocessing_wrapper():
    pass


def get_pcrs_postureinfo_for_one_policy_with_multiprocessing(
        pcrs_hostids_file:Path,
        batch_queue_process: multiprocessing.Process,
        batch_queue_of_file_paths: multiprocessing.Queue,
        max_batch_size=300
):
    with etld_lib_config.pcrs_open_file_compression_method(pcrs_hostids_file, "rt", encoding='utf-8') as read_file:
        pcrs_dict = json.load(read_file)

    pcrs_hostids_file_name_split = str(pcrs_hostids_file).split('_')
    # Get Batch Number from File Name
    pcrs_hostids_file_name_split.reverse()

    batch_number_str = "0"
    while pcrs_hostids_file_name_split:
        component = pcrs_hostids_file_name_split.pop()
        if component == 'batch':
            batch_number_str = pcrs_hostids_file_name_split.pop()
            break
    batch_number_str = f'batch_{batch_number_str}'

    subscription_id =  pcrs_dict[0]['subscriptionId']
    policy_id =  pcrs_dict[0]['policyId']
    pcrs_hostids_template_str = \
        '[{' + f'"policyId": "{policy_id}","subscriptionId":"{subscription_id}","hostIds":[]' + '}]'

    pcrs_hostids_template_dict = json.loads(pcrs_hostids_template_str)
    hostids_slice_counter = 0
    hostids_batch_counter = 0
    # Extract Posture info
    etld_lib_functions.logger.info(f"start get posture info for policy_id: {policy_id}, File: {Path(pcrs_hostids_file).name}")
    count_total_hostids_for_policy_id = 0
    proc_list = []
    for hostid in list(pcrs_dict[0]['hostIds']):
        pcrs_hostids_template_dict[0]['hostIds'].append(hostid)
        hostids_batch_counter = hostids_batch_counter + 1
        count_total_hostids_for_policy_id = count_total_hostids_for_policy_id + 1

        if int(hostids_batch_counter) >= int(max_batch_size):
            hostids_batch_counter = 0
            hostids_slice_counter = hostids_slice_counter + 1
            batch_number_and_slice_and_policy_str = f'{batch_number_str}_slice_{hostids_slice_counter:06d}_policyId_{policy_id}'

            if etld_lib_config.pcrs_postureinfo_turn_on_multiprocessing_flag == True:
                etld_lib_functions.logger.info(f"spawn {batch_number_and_slice_and_policy_str}")
                proc_list = get_pcrs_postureinfo_for_batch_of_hostids_multiprocessing(
                    batch_number_str=batch_number_and_slice_and_policy_str,
                    policy_id=policy_id,
                    payload_postureinfo_dict=pcrs_hostids_template_dict,
                    batch_queue_process=batch_queue_process,
                    batch_queue_of_file_paths=batch_queue_of_file_paths,
                    proc_list=proc_list,
                    concurrency_limit=etld_lib_config.pcrs_concurrency_limit
                )
            else:
                get_pcrs_postureinfo_for_batch_of_hostids(batch_number_str=batch_number_and_slice_and_policy_str,
                                                          policy_id=policy_id,
                                                          payload_postureinfo_dict=pcrs_hostids_template_dict,
                                                          batch_queue_process=batch_queue_process,
                                                          batch_queue_of_file_paths=batch_queue_of_file_paths)

            pcrs_hostids_template_dict = json.loads(pcrs_hostids_template_str)

    # Final extract if needed.
    if len(pcrs_hostids_template_dict[0]['hostIds']) > 0:
        #print(f"{batch_number_str}: {pcrs_hostids_template_dict}")
        hostids_slice_counter = hostids_slice_counter + 1
        batch_number_and_slice_and_policy_str = f'{batch_number_str}_slice_{hostids_slice_counter:06d}_policyId_{policy_id}'
        get_pcrs_postureinfo_for_batch_of_hostids(batch_number_str=batch_number_and_slice_and_policy_str,
                                                   policy_id=policy_id,
                                                   payload_postureinfo_dict=pcrs_hostids_template_dict,
                                                   batch_queue_process=batch_queue_process,
                                                   batch_queue_of_file_paths=batch_queue_of_file_paths)

    if etld_lib_config.pcrs_postureinfo_turn_on_multiprocessing_flag == True:
        etld_lib_functions.logger.info(f"Start Block waiting for final pcrs_postureinfo policy_id: {policy_id} processes to finish.")
        block_waiting_for_final_processes_to_finish(proc_list)
        validate_processes_finished_exit_code_zero(proc_list, reporting=True)
        etld_lib_functions.logger.info(f"End Block waiting for final pcrs_postureinfo policy_id: {policy_id} processes to finish.")

    etld_lib_functions.logger.info(f"end   get posture info for policy_id: {policy_id}, count: {count_total_hostids_for_policy_id:,}")



def get_pcrs_postureinfo_for_batch_of_hostids_multiprocessing(
        batch_number_str,
        policy_id,
        payload_postureinfo_dict,
        batch_queue_process,
        batch_queue_of_file_paths,
        proc_list,
        concurrency_limit
):

    pcrs_postureinfo_arg_dict = {
        'batch_number_str': batch_number_str,
        'policy_id': policy_id,
        'payload_postureinfo_dict': payload_postureinfo_dict,
        'batch_queue_process': batch_queue_process,
        'batch_queue_of_file_paths': batch_queue_of_file_paths,
        'api_call_name': batch_number_str
    }
    pcrs_postureinfo_proc = multiprocessing.Process(
        target=get_pcrs_postureinfo_for_batch_of_hostids_multiprocessing_wrapper,
        name=batch_number_str,
        args=(pcrs_postureinfo_arg_dict,)
    )
    pcrs_postureinfo_proc.daemon = True
    pcrs_postureinfo_proc.start()
    proc_list.append(pcrs_postureinfo_proc)
    block_if_max_concurrent_processes_running(concurrency_limit, proc_list)
    return proc_list

def get_pcrs_postureinfo_for_batch_of_hostids_multiprocessing_wrapper(pcrs_postureinfo_arg_dict):

    get_pcrs_postureinfo_for_batch_of_hostids(pcrs_postureinfo_arg_dict['batch_number_str'],
                                              pcrs_postureinfo_arg_dict['policy_id'],
                                              pcrs_postureinfo_arg_dict['payload_postureinfo_dict'],
                                              pcrs_postureinfo_arg_dict['batch_queue_process'],
                                              pcrs_postureinfo_arg_dict['batch_queue_of_file_paths'])


def get_pcrs_postureinfo_for_batch_of_hostids(
        batch_number_str, policy_id, payload_postureinfo_dict, batch_queue_process,batch_queue_of_file_paths):

    qualys_headers_dict = {}

    file_info_dict = \
        etld_lib_config.prepare_extract_batch_file_name(
            next_batch_number_str=batch_number_str,
            next_batch_date=utc_datetime,
            extract_dir=etld_lib_config.pcrs_extract_dir,
            file_name_type="pcrs_postureinfo",
            file_name_option="lastEvaluationDate",
            file_name_option_date=etld_lib_config.pcrs_last_evaluation_date,
            file_extension="json",
            compression_method=etld_lib_config.pcrs_open_file_compression_method)

    etld_lib_authentication_objects.qualys_authentication_obj.get_current_bearer_token()
    cred_dict = etld_lib_authentication_objects.qualys_authentication_obj.get_credentials_dict()

    # posture_info_dict = pcrs_04_extract_from_qualys.pcrs_extract_postureinfo(
    posture_info_chunk_of_characters=pcrs_04_extract_from_qualys.pcrs_extract_postureinfo(
        last_scan_date=etld_lib_config.pcrs_last_scan_date,
        last_evaluation_date=etld_lib_config.pcrs_last_evaluation_date,
        batch_number_str=batch_number_str,
        qualys_headers_dict=qualys_headers_dict,
        cred_dict=cred_dict,
        policy_id=policy_id,
        payload_postureinfo_dict=payload_postureinfo_dict,
        payload_dict=etld_lib_config.pcrs_postureinfo_payload_option,
        file_info_dict=file_info_dict)

    # Json usually starts with '[ {"id":...'
    if '{"' in posture_info_chunk_of_characters:
        add_batch_to_transform_json_to_sqlite(
            batch_number_str=batch_number_str,
            json_file=file_info_dict['next_file_path'],
            batch_queue_process=batch_queue_process,
            batch_queue_of_file_paths=batch_queue_of_file_paths)
    else:
        etld_lib_functions.logger.info(f"PolicyID: {policy_id}, Zero Hosts Found Active")


def block_if_max_concurrent_processes_running(max_proc_count, proc_list):
    proc_item: multiprocessing.Process
    while True:
        is_alive_count = 0
        for proc_item in proc_list:
            if proc_item.is_alive():
               is_alive_count = is_alive_count + 1
        if int(is_alive_count) >= int(max_proc_count):
           time.sleep(1)
           validate_processes_finished_exit_code_zero(proc_list)
        else:
            break

def block_waiting_for_final_processes_to_finish(proc_list):
    proc_item: multiprocessing.Process
    while True:
        is_alive_count = 0
        for proc_item in proc_list:
            if proc_item.is_alive():
                is_alive_count = is_alive_count + 1
        if is_alive_count >0:
            time.sleep(1)
        else:
            break

def validate_processes_finished_exit_code_zero(proc_list, reporting=False):
    for proc_item in proc_list:
        if proc_item.is_alive():
            pass
        elif proc_item.exitcode == 0:
            if reporting:
                etld_lib_functions.logger.info(f"process info: {proc_item}")
        else:
            etld_lib_functions.logger.error(f"{proc_item.name} exited with error code: {proc_item.exitcode}")
            etld_lib_functions.logger.error(f"{proc_item.name} failed, please investigate why job ran so long and rerun.")
            for proc_item_to_kill in proc_list:
                proc_item_to_kill.kill()
            exit(1)

def test_system_option(file_path: Path, has_more_records, number_of_files_to_extract, test_system_flag):
    if test_system_flag is True:
        test_batch_number = \
            etld_lib_extract_transform_load.get_batch_number_from_filename(file_path)
        if int(test_batch_number) >= int(number_of_files_to_extract):
            has_more_records = '0'
    return has_more_records


def main():
    global utc_datetime
    utc_datetime = etld_lib_datetime.get_utc_datetime_qualys_format()
    begin_pcrs_03_extract_controller()
    remove_old_files()
    batch_queue_of_file_paths, batch_queue_process = start_multiprocessing_transform_json_to_sqlite()
    try:
        get_pcrs_policy_list_data(batch_queue_process=batch_queue_process,
                                  batch_queue_of_file_paths=batch_queue_of_file_paths)
        get_pcrs_hostids_data(batch_queue_process=batch_queue_process,
                              batch_queue_of_file_paths=batch_queue_of_file_paths)
        get_pcrs_postureinfo_data(batch_queue_process=batch_queue_process,
                                  batch_queue_of_file_paths=batch_queue_of_file_paths)
    except Exception as e:
        etld_lib_functions.logger.error(f"ERROR: get_pcrs function failed. policy_list or hostids_data or postureinfo_data.  Check log for last exec.")
        etld_lib_functions.logger.error(f"ERROR: {e}")
        formatted_lines = traceback.format_exc().splitlines()
        etld_lib_functions.logger.error(f"ERROR: {formatted_lines}")
        exit(1)
    finally:
        stop_multiprocessing_transform_json_to_sqlite(batch_queue_of_file_paths=batch_queue_of_file_paths,
                                                      batch_queue_process=batch_queue_process)

    end_pcrs_03_extract_controller()


if __name__ == "__main__":
    etld_lib_functions.main(my_logger_prog_name='pcrs_03_extract_controller')
    etld_lib_config.main()
    etld_lib_authentication_objects.main()
    main()



